"""Helpers for the parser module."""

from typing import Tuple, List, Any, Iterator, TYPE_CHECKING

from sqlfluff.core.errors import SQLParseError
from sqlfluff.core.string_helpers import curtail_string

if TYPE_CHECKING:
    from sqlfluff.core.parser.segments import BaseSegment  # pragma: no cover


def join_segments_raw(segments: Tuple["BaseSegment", ...]) -> str:
    """Make a string from the joined `raw` attributes of an iterable of segments."""
    return "".join(s.raw for s in segments)


def join_segments_raw_curtailed(segments: Tuple["BaseSegment", ...], length=20) -> str:
    """Make a string up to a certain length from an iterable of segments."""
    return curtail_string(join_segments_raw(segments), length=length)


def check_still_complete(
    segments_in: Tuple["BaseSegment", ...],
    matched_segments: Tuple["BaseSegment", ...],
    unmatched_segments: Tuple["BaseSegment", ...],
) -> bool:
    """Check that the segments in are the same as the segments out."""
    initial_str = join_segments_raw(segments_in)
    current_str = join_segments_raw(matched_segments + unmatched_segments)

    if initial_str != current_str:  # pragma: no cover
        segment = unmatched_segments[0] if unmatched_segments else None
        raise SQLParseError(
            f"Could not parse: {current_str}",
            segment=segment,
        )
    return True


def trim_non_code_segments(
    segments: Tuple["BaseSegment", ...]
) -> Tuple[
    Tuple["BaseSegment", ...], Tuple["BaseSegment", ...], Tuple["BaseSegment", ...]
]:
    """Take segments and split off surrounding non-code segments as appropriate.

    We use slices to avoid creating too many unnecessary tuples.
    """
    pre_idx = 0
    seg_len = len(segments)
    post_idx = seg_len

    if segments:
        seg_len = len(segments)

        # Trim the start
        while pre_idx < seg_len and not segments[pre_idx].is_code:
            pre_idx += 1

        # Trim the end
        while post_idx > pre_idx and not segments[post_idx - 1].is_code:
            post_idx -= 1

    return segments[:pre_idx], segments[pre_idx:post_idx], segments[post_idx:]


def iter_indices(seq: List, val: Any) -> Iterator[int]:
    """Iterate all indices in a list that val occurs at.

    Args:
        seq (list): A list to look for indices in.
        val: What to look for.

    Yields:
        int: The index of val in seq.

    Examples:
        The function works like str.index() but iterates all
        the results rather than returning the first.

        >>> print([i for i in iter_indices([1, 0, 2, 3, 2], 2)])
        [2, 4]
    """
    for idx, el in enumerate(seq):
        if el == val:
            yield idx



"""Script used for dbt templater tests."""
import os
import sys
from typing import Optional, Tuple

from jinja2.sandbox import SandboxedEnvironment

DBT_VERSION_TUPLE: Optional[Tuple[int, int]] = None
try:
    from sqlfluff_templater_dbt.templater import DBT_VERSION_TUPLE
except ImportError:
    pass


def main(project_dir):
    """Load Jinja template file expand it write to packages.yml."""
    env = SandboxedEnvironment()
    with open(os.path.join(project_dir, "packages.yml.jinja2")) as f:
        template_str = f.read()
    template = env.from_string(
        template_str, globals=dict(DBT_VERSION_TUPLE=DBT_VERSION_TUPLE)
    )
    expanded = template.render()
    with open(os.path.join(project_dir, "packages.yml"), "w") as f:
        f.write(expanded)


if __name__ == "__main__":
    if DBT_VERSION_TUPLE is not None:
        main(sys.argv[1])



"""Tests the python routines within L048."""
import sqlfluff


def test__rules__std_L048_raised() -> None:
    """L048 is raised for quoted literals not surrounded by a single whitespace."""
    sql = "SELECT a +'b'+'c' FROM tbl;"
    result = sqlfluff.lint(sql)

    assert len(result) == 7

    results_l048 = [r for r in result if r["code"] == "L048"]
    assert len(results_l048) == 3
    assert (
        results_l048[0]["description"]
        == "Expected single whitespace between binary operator '+' and quoted literal."
    )
    assert (
        results_l048[1]["description"]
        == "Expected single whitespace between quoted literal and binary operator '+'."
    )
    assert (
        results_l048[2]["description"]
        == "Expected single whitespace between binary operator '+' and quoted literal."
    )



"""Implementation of Rule L044."""
from typing import Optional

from sqlfluff.utils.analysis.select_crawler import Query, SelectCrawler
from sqlfluff.core.parser import BaseSegment
from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.doc_decorators import document_groups
from sqlfluff.utils.functional import sp, FunctionalContext


_START_TYPES = ["select_statement", "set_expression", "with_compound_statement"]


class RuleFailure(Exception):
    """Exception class for reporting lint failure inside deeply nested code."""

    def __init__(self, anchor: BaseSegment):
        self.anchor: BaseSegment = anchor


@document_groups
class Rule_L044(BaseRule):
    """Query produces an unknown number of result columns.

    **Anti-pattern**

    Querying all columns using ``*`` produces a query result where the number
    or ordering of columns changes if the upstream table's schema changes.
    This should generally be avoided because it can cause slow performance,
    cause important schema changes to go undetected, or break production code.
    For example:

    * If a query does ``SELECT t.*`` and is expected to return columns ``a``, ``b``,
      and ``c``, the actual columns returned will be wrong/different if columns
      are added to or deleted from the input table.
    * ``UNION`` and ``DIFFERENCE`` clauses require the inputs have the same number
      of columns (and compatible types).
    * ``JOIN`` queries may break due to new column name conflicts, e.g. the
      query references a column ``c`` which initially existed in only one input
      table but a column of the same name is added to another table.
    * ``CREATE TABLE (<<column schema>>) AS SELECT *``


    .. code-block:: sql

        WITH cte AS (
            SELECT * FROM foo
        )

        SELECT * FROM cte
        UNION
        SELECT a, b FROM t

    **Best practice**

    Somewhere along the "path" to the source data, specify columns explicitly.

    .. code-block:: sql

        WITH cte AS (
            SELECT * FROM foo
        )

        SELECT a, b FROM cte
        UNION
        SELECT a, b FROM t

    """

    groups = ("all",)
    crawl_behaviour = SegmentSeekerCrawler(set(_START_TYPES))

    def _handle_alias(self, selectable, alias_info, query):
        select_info_target = SelectCrawler.get(
            query, alias_info.from_expression_element
        )[0]
        if isinstance(select_info_target, str):
            # It's an alias to an external table whose
            # number of columns could vary without our
            # knowledge. Thus, warn.
            self.logger.debug(
                f"Query target {select_info_target} is external. Generating warning."
            )
            raise RuleFailure(selectable.selectable)
        else:
            # Handle nested SELECT.
            self._analyze_result_columns(select_info_target)

    def _analyze_result_columns(self, query: Query):
        """Given info on a list of SELECTs, determine whether to warn."""
        # Recursively walk from the given query (select_info_list) to any
        # wildcard columns in the select targets. If every wildcard evdentually
        # resolves to a query without wildcards, all is well. Otherwise, warn.
        if not query.selectables:
            return  # pragma: no cover
        for selectable in query.selectables:
            self.logger.debug(f"Analyzing query: {selectable.selectable.raw}")
            for wildcard in selectable.get_wildcard_info():
                if wildcard.tables:
                    for wildcard_table in wildcard.tables:
                        self.logger.debug(
                            f"Wildcard: {wildcard.segment.raw} has target "
                            "{wildcard_table}"
                        )
                        # Is it an alias?
                        alias_info = selectable.find_alias(wildcard_table)
                        if alias_info:
                            # Found the alias matching the wildcard. Recurse,
                            # analyzing the query associated with that alias.
                            self._handle_alias(selectable, alias_info, query)
                        else:
                            # Not an alias. Is it a CTE?
                            cte = query.lookup_cte(wildcard_table)
                            if cte:
                                # Wildcard refers to a CTE. Analyze it.
                                self._analyze_result_columns(cte)
                            else:
                                # Not CTE, not table alias. Presumably an
                                # external table. Warn.
                                self.logger.debug(
                                    f"Query target {wildcard_table} is external. "
                                    "Generating warning."
                                )
                                raise RuleFailure(selectable.selectable)
                else:
                    # No table was specified with the wildcard. Assume we're
                    # querying from a nested select in FROM.
                    query_list = SelectCrawler.get(
                        query, query.selectables[0].selectable
                    )
                    for o in query_list:
                        if isinstance(o, Query):
                            self._analyze_result_columns(o)
                            return
                    self.logger.debug(
                        f'Query target "{query.selectables[0].selectable.raw}" has no '
                        "targets. Generating warning."
                    )
                    raise RuleFailure(query.selectables[0].selectable)

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Outermost query should produce known number of columns."""
        if not FunctionalContext(context).parent_stack.any(sp.is_type(*_START_TYPES)):
            crawler = SelectCrawler(context.segment, context.dialect)

            # Begin analysis at the outer query.
            if crawler.query_tree:
                try:
                    return self._analyze_result_columns(crawler.query_tree)
                except RuleFailure as e:
                    return LintResult(anchor=e.anchor)
        return None



"""Implementation of Rule L025."""

from dataclasses import dataclass, field
from typing import cast, List, Set

from sqlfluff.core.dialects.base import Dialect
from sqlfluff.core.parser.segments import BaseSegment
from sqlfluff.utils.analysis.select import get_select_statement_info
from sqlfluff.utils.analysis.select_crawler import (
    Query as SelectCrawlerQuery,
    SelectCrawler,
)
from sqlfluff.core.rules import (
    BaseRule,
    LintFix,
    LintResult,
    RuleContext,
    EvalResultType,
)
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.doc_decorators import document_fix_compatible, document_groups
from sqlfluff.utils.functional import Segments, sp
from sqlfluff.core.dialects.common import AliasInfo


@dataclass
class L025Query(SelectCrawlerQuery):
    """SelectCrawler Query with custom L025 info."""

    aliases: List[AliasInfo] = field(default_factory=list)
    tbl_refs: Set[str] = field(default_factory=set)


@document_groups
@document_fix_compatible
class Rule_L025(BaseRule):
    """Tables should not be aliased if that alias is not used.

    **Anti-pattern**

    .. code-block:: sql

        SELECT
            a
        FROM foo AS zoo

    **Best practice**

    Use the alias or remove it. An unused alias makes code
    harder to read without changing any functionality.

    .. code-block:: sql

        SELECT
            zoo.a
        FROM foo AS zoo

        -- Alternatively...

        SELECT
            a
        FROM foo

    """

    groups = ("all", "core")
    crawl_behaviour = SegmentSeekerCrawler({"select_statement"})
    _dialects_requiring_alias_for_values_clause = [
        "snowflake",
        "tsql",
    ]

    def _eval(self, context: RuleContext) -> EvalResultType:
        violations: List[LintResult] = []
        assert context.segment.is_type("select_statement")
        # Exit early if the SELECT does not define any aliases.
        select_info = get_select_statement_info(context.segment, context.dialect)
        if not select_info or not select_info.table_aliases:
            return None

        # Analyze the SELECT.
        crawler = SelectCrawler(context.segment, context.dialect, query_class=L025Query)
        query: L025Query = cast(L025Query, crawler.query_tree)
        self._analyze_table_aliases(query, context.dialect)

        alias: AliasInfo
        for alias in query.aliases:
            # Skip alias if it's required (some dialects require aliases for
            # VALUES clauses).
            if alias.from_expression_element and self.is_alias_required(
                alias.from_expression_element, context.dialect.name
            ):
                continue

            if alias.aliased and alias.ref_str not in query.tbl_refs:
                # Unused alias. Report and fix.
                violations.append(self._report_unused_alias(alias))
        return violations or None

    @classmethod
    def is_alias_required(
        cls, from_expression_element: BaseSegment, dialect_name: str
    ) -> bool:
        """Given an alias, is it REQUIRED to be present?

        Aliases are required in SOME, but not all dialects when there's a VALUES
        clause.
        """
        # Look for a table_expression (i.e. VALUES clause) as a descendant of
        # the FROM expression, potentially nested inside brackets. The reason we
        # allow nesting in brackets is that in some dialects (e.g. TSQL), this
        # is actually *required* in order for SQL Server to parse it.
        for segment in from_expression_element.iter_segments(expanding=("bracketed",)):
            if segment.is_type("table_expression"):
                # Found a table expression. Does it have a VALUES clause?
                if segment.get_child("values_clause"):
                    # Found a VALUES clause. Is this a dialect that requires
                    # VALUE clauses to be aliased?
                    return (
                        dialect_name in cls._dialects_requiring_alias_for_values_clause
                    )
                elif any(
                    seg.is_type(
                        "select_statement", "set_expression", "with_compound_statement"
                    )
                    for seg in segment.iter_segments(expanding=("bracketed",))
                ):
                    # The FROM expression is a derived table, i.e. a nested
                    # SELECT. In this case, the alias is required in every
                    # dialect we checked (MySQL, Postgres, T-SQL).
                    # https://pganalyze.com/docs/log-insights/app-errors/U115
                    return True
                else:
                    # None of the special cases above applies, so the alias is
                    # not required.
                    return False

        # This should never happen. Return False just to be safe.
        return False  # pragma: no cover

    @classmethod
    def _analyze_table_aliases(cls, query: L025Query, dialect: Dialect):
        # Get table aliases defined in query.
        for selectable in query.selectables:
            select_info = selectable.select_info
            if select_info:
                # Record the aliases.
                query.aliases += select_info.table_aliases

                # Look at each table reference; if it's an alias reference,
                # resolve the alias: could be an alias defined in "query"
                # itself or an "ancestor" query.
                for r in select_info.reference_buffer:
                    for tr in r.extract_possible_references(
                        level=r.ObjectReferenceLevel.TABLE
                    ):
                        # This function walks up the query's parent stack if necessary.
                        cls._resolve_and_mark_reference(query, tr.part)

        # Visit children.
        for child in query.children:
            cls._analyze_table_aliases(cast(L025Query, child), dialect)

    @classmethod
    def _resolve_and_mark_reference(cls, query: L025Query, ref: str):
        # Does this query define the referenced alias?
        if any(ref == a.ref_str for a in query.aliases):
            # Yes. Record the reference.
            query.tbl_refs.add(ref)
        elif query.parent:
            # No. Recursively check the query's parent hierarchy.
            cls._resolve_and_mark_reference(cast(L025Query, query.parent), ref)

    @classmethod
    def _report_unused_alias(cls, alias: AliasInfo) -> LintResult:
        fixes = [LintFix.delete(alias.alias_expression)]  # type: ignore
        # Walk back to remove indents/whitespaces
        to_delete = (
            Segments(*alias.from_expression_element.segments)
            .reversed()
            .select(
                start_seg=alias.alias_expression,
                # Stop once we reach an other, "regular" segment.
                loop_while=sp.or_(sp.is_whitespace(), sp.is_meta()),
            )
        )
        fixes += [LintFix.delete(seg) for seg in to_delete]
        return LintResult(
            anchor=alias.segment,
            description="Alias {!r} is never used in SELECT statement.".format(
                alias.ref_str
            ),
            fixes=fixes,
        )



"""Allows to provide deprecated options for click's command."""

from typing import Any, Callable

import click
from click import Context, OptionParser, echo, style
from click.parser import Option, ParsingState


class DeprecatedOption(click.Option):
    """Allows to provide deprecated options for click's command.

    Works with `DeprecatedOptionsCommand` (see below).
    Expects to be provided into standard `@click.option` with:
      * two parameter declarations arguments - old one (deprecated)
        and new one (preferred);
      * `cls` parameter (standard click Option) as `cls=DeprecatedOption`;
      * `deprecated` parameter - which says which ones are deprecated,
        like`deprecated=["--disable_progress_bar"]1.

    This is based on
      * https://stackoverflow.com/a/50402799/5172513

    It's somewhat hackish and may broke when click internals are changed, it is even
    mentioned in SO:
    > This code reaches into some private structures in the parser, but this is
    unlikely to be an issue. This parser code was last changed 4 years ago.
    The parser code is unlikely to undergo significant revisions.

    Hopefully will be removed when
      * https://github.com/pallets/click/issues/2263
    is finished.
    """

    def __init__(self, *args, **kwargs):
        self.deprecated = kwargs.pop("deprecated", ())
        self.preferred = args[0][-1]

        super().__init__(*args, **kwargs)


class DeprecatedOptionsCommand(click.Command):
    """Allows to provide deprecated options for click's command.

    Works with `DeprecatedOption` (see above).
    Expects to be provided into standard `@click.command` as:
      * `@cli.command(cls=DeprecatedOptionsCommand)`
    """

    def make_parser(self, ctx: Context) -> OptionParser:
        """Hook 'make_parser' and during processing check the name.

        Used to invoke the option to see if it is preferred.
        """
        parser: OptionParser = super().make_parser(ctx)

        # get the parser options
        options = set(parser._short_opt.values())
        options |= set(parser._long_opt.values())

        for option in options:
            if not isinstance(option.obj, DeprecatedOption):
                continue

            option.process = self._make_process(option)  # type: ignore

        return parser

    def _make_process(self, an_option: Option) -> Callable:
        """Construct a closure to the parser option processor."""
        orig_process: Callable = an_option.process
        deprecated = getattr(an_option.obj, "deprecated", None)
        preferred = getattr(an_option.obj, "preferred", None)

        if not deprecated:
            raise ValueError(
                f"Expected `deprecated` value for `{an_option.obj.name!r}`"
            )

        def process(value: Any, state: ParsingState) -> None:
            """Custom process method.

            The function above us on the stack used 'opt' to
            pick option from a dict, see if it is deprecated.
            """
            # reach up the stack and get 'opt'
            import inspect

            frame = inspect.currentframe()
            try:
                opt = frame.f_back.f_locals.get("opt")  # type: ignore
            finally:
                del frame

            if opt in deprecated:  # type: ignore
                msg = (
                    f"DeprecationWarning: The option {opt!r} is deprecated, "
                    f"use {preferred!r}."
                )
                echo(style(msg, fg="red"), err=True)

            return orig_process(value, state)

        return process



"""Defines the hook endpoints for the dbt templater plugin."""

from sqlfluff_templater_dbt.templater import DbtTemplater
from sqlfluff.core.plugin import hookimpl


@hookimpl
def get_templaters():
    """Get templaters."""
    return [DbtTemplater]



"""Implementation of Rule L065."""
from typing import List

from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.doc_decorators import document_fix_compatible, document_groups
from sqlfluff.utils.reflow.sequence import ReflowSequence


@document_groups
@document_fix_compatible
class Rule_L065(BaseRule):
    """Set operators should be surrounded by newlines.

    **Anti-pattern**

    In this example, `UNION ALL` is not on a line itself.

    .. code-block:: sql

        SELECT 'a' AS col UNION ALL
        SELECT 'b' AS col

    **Best practice**

    .. code-block:: sql

        SELECT 'a' AS col
        UNION ALL
        SELECT 'b' AS col

    """

    groups = ("all",)

    crawl_behaviour = SegmentSeekerCrawler({"set_operator"})

    def _eval(self, context: RuleContext) -> List[LintResult]:
        """Set operators should be surrounded by newlines.

        For any set operator we check if there is any NewLineSegment in the non-code
        segments preceding or following it.

        In particular, as part of this rule we allow multiple NewLineSegments.
        """
        return (
            ReflowSequence.from_around_target(
                context.segment,
                root_segment=context.parent_stack[0],
                config=context.config,
            )
            .rebreak()
            .get_results()
        )



"""Keywords in the Postgres Dialect.

Most of the keywords come from
https://www.postgresql.org/docs/13/sql-keywords-appendix.html
Here, "not-keyword" refers to a word not being a keyword, and will be removed from any
default keyword definition, these keywords are, or have been, an ANSI keyword.

There are also some keywords that are(n't) supported as types and function, but there
isn't support for that distinction at present.
"""


def priority_keyword_merge(*args):
    """Merge keyword lists, giving priority to entries in later lists.

    *args is a list of keyword lists, these lists should be of tuples in the form
    (keyword, type)

    """
    keyword_lists = [*args]
    base_list = []
    if len(keyword_lists) == 1:
        return keyword_lists[0]

    while len(keyword_lists) > 1:
        base_list, priority_list = keyword_lists[0], keyword_lists[1]
        keyword_set = set([x[0] for x in base_list])
        for item in priority_list:
            if item[0] in keyword_set:
                for index, keyword in enumerate(base_list):
                    if keyword[0] == item[0]:
                        base_list.pop(index)
                        break
            base_list.append(item)

        keyword_lists.pop(1)

    return base_list


def get_keywords(keyword_list, keyword_type):
    """Get a list of keywords of the required type.

    keyword_type should be one of "not-keyword", "reserved", "non-reserved"
    """
    keywords = [x[0] for x in keyword_list if x[1].startswith(keyword_type)]

    return keywords


postgres_docs_keywords = [
    ("A", "not-keyword"),
    ("ABORT", "non-reserved"),
    ("ABS", "not-keyword"),
    ("ABSENT", "not-keyword"),
    ("ABSOLUTE", "non-reserved"),
    ("ACCESS", "non-reserved"),
    ("ACCORDING", "not-keyword"),
    ("ACOS", "not-keyword"),
    ("ACTION", "non-reserved"),
    ("ADA", "not-keyword"),
    ("ADD", "non-reserved"),
    ("ADMIN", "non-reserved"),
    ("AFTER", "non-reserved"),
    ("AGGREGATE", "non-reserved"),
    ("ALL", "reserved"),
    ("ALLOCATE", "not-keyword"),
    ("ALSO", "non-reserved"),
    ("ALTER", "non-reserved"),
    ("ALWAYS", "non-reserved"),
    ("ANALYSE", "reserved"),
    ("ANALYZE", "reserved"),
    ("AND", "reserved"),
    ("ANY", "reserved"),
    ("ARE", "not-keyword"),
    ("ARRAY", "reserved"),
    ("ARRAY_AGG", "not-keyword"),
    ("ARRAY_MAX_CARDINALITY", "not-keyword"),
    ("AS", "reserved"),
    ("ASC", "reserved"),
    ("ASENSITIVE", "not-keyword"),
    ("ASIN", "not-keyword"),
    ("ASSERTION", "non-reserved"),
    ("ASSIGNMENT", "non-reserved"),
    ("ASYMMETRIC", "reserved"),
    ("AT", "non-reserved"),
    ("ATAN", "not-keyword"),
    ("ATOMIC", "not-keyword"),
    ("ATTACH", "non-reserved"),
    ("ATTRIBUTE", "non-reserved"),
    ("ATTRIBUTES", "not-keyword"),
    ("AUTHORIZATION", "reserved-(can-be-function-or-type)"),
    ("AVG", "not-keyword"),
    ("BACKWARD", "non-reserved"),
    ("BASE64", "not-keyword"),
    ("BEFORE", "non-reserved"),
    ("BEGIN", "non-reserved"),
    ("BEGIN_FRAME", "not-keyword"),
    ("BEGIN_PARTITION", "not-keyword"),
    ("BERNOULLI", "not-keyword"),
    ("BETWEEN", "non-reserved-(cannot-be-function-or-type)"),
    ("BIGINT", "non-reserved-(cannot-be-function-or-type)"),
    ("BIGSERIAL", "non-reserved-(cannot-be-function-or-type)"),
    ("BINARY", "reserved-(can-be-function-or-type)"),
    ("BIT", "non-reserved-(cannot-be-function-or-type)"),
    ("BIT_LENGTH", "not-keyword"),
    ("BLOB", "not-keyword"),
    ("BLOCKED", "not-keyword"),
    ("BOM", "not-keyword"),
    ("BOOLEAN", "non-reserved-(cannot-be-function-or-type)"),
    ("BOOL", "non-reserved-(cannot-be-function-or-type)"),
    ("BOTH", "reserved"),
    ("BOX", "non-reserved-(cannot-be-function-or-type)"),
    ("BREADTH", "not-keyword"),
    ("BY", "non-reserved"),
    ("BYTEA", "non-reserved-(cannot-be-function-or-type)"),
    ("C", "not-keyword"),
    ("CACHE", "non-reserved"),
    ("CALL", "non-reserved"),
    ("CALLED", "non-reserved"),
    ("CARDINALITY", "not-keyword"),
    ("CASCADE", "non-reserved"),
    ("CASCADED", "non-reserved"),
    ("CASE", "reserved"),
    ("CAST", "reserved"),
    ("CATALOG", "non-reserved"),
    ("CATALOG_NAME", "not-keyword"),
    ("CEIL", "not-keyword"),
    ("CEILING", "not-keyword"),
    ("CHAIN", "non-reserved"),
    ("CHAINING", "not-keyword"),
    ("CHAR", "non-reserved-(cannot-be-function-or-type)"),
    ("CHARACTER", "non-reserved-(cannot-be-function-or-type)"),
    ("CHARACTERISTICS", "non-reserved"),
    ("CHARACTERS", "not-keyword"),
    ("CHARACTER_LENGTH", "not-keyword"),
    ("CHARACTER_SET_CATALOG", "not-keyword"),
    ("CHARACTER_SET_NAME", "not-keyword"),
    ("CHARACTER_SET_SCHEMA", "not-keyword"),
    ("CHAR_LENGTH", "not-keyword"),
    ("CHECK", "reserved"),
    ("CHECKPOINT", "non-reserved"),
    ("CIDR", "non-reserved-(cannot-be-function-or-type)"),
    ("CIRCLE", "non-reserved-(cannot-be-function-or-type)"),
    ("CLASS", "non-reserved"),
    ("CLASSIFIER", "not-keyword"),
    ("CLASS_ORIGIN", "not-keyword"),
    ("CLOB", "not-keyword"),
    ("CLOSE", "non-reserved"),
    ("CLUSTER", "non-reserved"),
    ("COALESCE", "non-reserved-(cannot-be-function-or-type)"),
    ("COBOL", "not-keyword"),
    ("COLLATE", "reserved"),
    ("COLLATION", "reserved-(can-be-function-or-type)"),
    ("COLLATION_CATALOG", "not-keyword"),
    ("COLLATION_NAME", "not-keyword"),
    ("COLLATION_SCHEMA", "not-keyword"),
    ("COLLECT", "not-keyword"),
    ("COLUMN", "reserved"),
    ("COLUMNS", "non-reserved"),
    ("COLUMN_NAME", "not-keyword"),
    ("COMMAND_FUNCTION", "not-keyword"),
    ("COMMAND_FUNCTION_CODE", "not-keyword"),
    ("COMMENT", "non-reserved"),
    ("COMMENTS", "non-reserved"),
    ("COMMIT", "non-reserved"),
    ("COMMITTED", "non-reserved"),
    ("COMPRESSION", "non-reserved"),
    ("CONCURRENTLY", "reserved-(can-be-function-or-type)"),
    ("CONDITION", "not-keyword"),
    ("CONDITIONAL", "not-keyword"),
    ("CONDITION_NUMBER", "not-keyword"),
    ("CONFIGURATION", "non-reserved"),
    ("CONFLICT", "non-reserved"),
    ("CONNECT", "not-keyword"),
    ("CONNECTION", "non-reserved"),
    ("CONNECTION_NAME", "not-keyword"),
    ("CONSTRAINT", "reserved"),
    ("CONSTRAINTS", "non-reserved"),
    ("CONSTRAINT_CATALOG", "not-keyword"),
    ("CONSTRAINT_NAME", "not-keyword"),
    ("CONSTRAINT_SCHEMA", "not-keyword"),
    ("CONSTRUCTOR", "not-keyword"),
    ("CONTAINS", "not-keyword"),
    ("CONTENT", "non-reserved"),
    ("CONTINUE", "non-reserved"),
    ("CONTROL", "not-keyword"),
    ("CONVERSION", "non-reserved"),
    ("CONVERT", "not-keyword"),
    ("COPY", "non-reserved"),
    ("CORR", "not-keyword"),
    ("CORRESPONDING", "not-keyword"),
    ("COS", "not-keyword"),
    ("COSH", "not-keyword"),
    ("COST", "non-reserved"),
    ("COUNT", "not-keyword"),
    ("COVAR_POP", "not-keyword"),
    ("COVAR_SAMP", "not-keyword"),
    ("CREATE", "reserved"),
    ("CROSS", "reserved-(can-be-function-or-type)"),
    ("CSV", "non-reserved"),
    ("CUBE", "non-reserved"),
    ("CUME_DIST", "not-keyword"),
    ("CURRENT", "non-reserved"),
    ("CURRENT_CATALOG", "reserved"),
    ("CURRENT_DATE", "reserved"),
    ("CURRENT_DEFAULT_TRANSFORM_GROUP", "not-keyword"),
    ("CURRENT_PATH", "not-keyword"),
    ("CURRENT_ROLE", "reserved"),
    ("CURRENT_ROW", "not-keyword"),
    ("CURRENT_SCHEMA", "reserved-(can-be-function-or-type)"),
    ("CURRENT_TIME", "reserved"),
    ("CURRENT_TIMESTAMP", "reserved"),
    ("CURRENT_TRANSFORM_GROUP_FOR_TYPE", "not-keyword"),
    ("CURRENT_USER", "reserved"),
    ("CURSOR", "non-reserved"),
    ("CURSOR_NAME", "not-keyword"),
    ("CYCLE", "non-reserved"),
    ("DATA", "non-reserved"),
    ("DATABASE", "non-reserved"),
    ("DATALINK", "not-keyword"),
    ("DATE", "not-keyword"),
    ("DATERANGE", "non-reserved-(cannot-be-function-or-type)"),
    ("DATETIME_INTERVAL_CODE", "not-keyword"),
    ("DATETIME_INTERVAL_PRECISION", "not-keyword"),
    ("DAY", "non-reserved"),
    ("DB", "not-keyword"),
    ("DEALLOCATE", "non-reserved"),
    ("DEC", "non-reserved-(cannot-be-function-or-type)"),
    ("DECFLOAT", "not-keyword"),
    ("DECIMAL", "non-reserved-(cannot-be-function-or-type)"),
    ("DECLARE", "non-reserved"),
    ("DEFAULT", "reserved"),
    ("DEFAULTS", "non-reserved"),
    ("DEFERRABLE", "reserved"),
    ("DEFERRED", "non-reserved"),
    ("DEFINE", "not-keyword"),
    ("DEFINED", "not-keyword"),
    ("DEFINER", "non-reserved"),
    ("DEGREE", "not-keyword"),
    ("DELETE", "non-reserved"),
    ("DELIMITER", "non-reserved"),
    ("DELIMITERS", "non-reserved"),
    ("DENSE_RANK", "not-keyword"),
    ("DEPENDS", "non-reserved"),
    ("DEPTH", "not-keyword"),
    ("DEREF", "not-keyword"),
    ("DERIVED", "not-keyword"),
    ("DESC", "reserved"),
    ("DESCRIBE", "not-keyword"),
    ("DESCRIPTOR", "not-keyword"),
    ("DETACH", "non-reserved"),
    ("DETERMINISTIC", "not-keyword"),
    ("DIAGNOSTICS", "not-keyword"),
    ("DICTIONARY", "non-reserved"),
    ("DISABLE", "non-reserved"),
    ("DISCARD", "non-reserved"),
    ("DISCONNECT", "not-keyword"),
    ("DISPATCH", "not-keyword"),
    ("DISTINCT", "reserved"),
    ("DLNEWCOPY", "not-keyword"),
    ("DLPREVIOUSCOPY", "not-keyword"),
    ("DLURLCOMPLETE", "not-keyword"),
    ("DLURLCOMPLETEONLY", "not-keyword"),
    ("DLURLCOMPLETEWRITE", "not-keyword"),
    ("DLURLPATH", "not-keyword"),
    ("DLURLPATHONLY", "not-keyword"),
    ("DLURLPATHWRITE", "not-keyword"),
    ("DLURLSCHEME", "not-keyword"),
    ("DLURLSERVER", "not-keyword"),
    ("DLVALUE", "not-keyword"),
    ("DO", "reserved"),
    ("DOCUMENT", "non-reserved"),
    ("DOMAIN", "non-reserved"),
    ("DOUBLE", "non-reserved"),
    ("DROP", "non-reserved"),
    ("DYNAMIC", "not-keyword"),
    ("DYNAMIC_FUNCTION", "not-keyword"),
    ("DYNAMIC_FUNCTION_CODE", "not-keyword"),
    ("EACH", "non-reserved"),
    ("ELEMENT", "not-keyword"),
    ("ELSE", "reserved"),
    ("EMPTY", "not-keyword"),
    ("ENABLE", "non-reserved"),
    ("ENCODING", "non-reserved"),
    ("ENCRYPTED", "non-reserved"),
    ("END", "reserved"),
    ("END-EXEC", "not-keyword"),
    ("END_FRAME", "not-keyword"),
    ("END_PARTITION", "not-keyword"),
    ("ENFORCED", "not-keyword"),
    ("ENUM", "non-reserved"),
    ("EQUALS", "not-keyword"),
    ("ERROR", "not-keyword"),
    ("ESCAPE", "non-reserved"),
    ("EVENT", "non-reserved"),
    ("EVERY", "not-keyword"),
    ("EXCEPT", "reserved"),
    ("EXCEPTION", "not-keyword"),
    ("EXCLUDE", "non-reserved"),
    ("EXCLUDING", "non-reserved"),
    ("EXCLUSIVE", "non-reserved"),
    ("EXEC", "not-keyword"),
    ("EXECUTE", "non-reserved"),
    ("EXISTS", "non-reserved-(cannot-be-function-or-type)"),
    ("EXP", "not-keyword"),
    ("EXPLAIN", "non-reserved"),
    ("EXPRESSION", "non-reserved"),
    ("EXTENSION", "non-reserved"),
    ("EXTERNAL", "non-reserved"),
    ("EXTRACT", "non-reserved-(cannot-be-function-or-type)"),
    ("FALSE", "reserved"),
    ("FAMILY", "non-reserved"),
    ("FETCH", "reserved"),
    ("FILE", "not-keyword"),
    ("FILTER", "non-reserved"),
    ("FINAL", "not-keyword"),
    ("FINALIZE", "non-reserved"),
    ("FINISH", "not-keyword"),
    ("FIRST", "non-reserved"),
    ("FIRST_VALUE", "not-keyword"),
    ("FLAG", "not-keyword"),
    ("FLOAT", "non-reserved-(cannot-be-function-or-type)"),
    ("FLOOR", "not-keyword"),
    ("FOLLOWING", "non-reserved"),
    ("FOR", "reserved"),
    ("FORCE", "non-reserved"),
    ("FOREIGN", "reserved"),
    ("FORMAT", "not-keyword"),
    ("FORTRAN", "not-keyword"),
    ("FORWARD", "non-reserved"),
    ("FOUND", "not-keyword"),
    ("FRAME_ROW", "not-keyword"),
    ("FREE", "not-keyword"),
    ("FREEZE", "reserved-(can-be-function-or-type)"),
    ("FROM", "reserved"),
    ("FS", "not-keyword"),
    ("FULFILL", "not-keyword"),
    ("FULL", "reserved-(can-be-function-or-type)"),
    ("FUNCTION", "non-reserved"),
    ("FUNCTIONS", "non-reserved"),
    ("FUSION", "not-keyword"),
    ("G", "not-keyword"),
    ("GENERAL", "not-keyword"),
    ("GENERATED", "non-reserved"),
    ("GET", "not-keyword"),
    ("GLOBAL", "non-reserved"),
    ("GO", "not-keyword"),
    ("GOTO", "not-keyword"),
    ("GRANT", "reserved"),
    ("GRANTED", "non-reserved"),
    ("GREATEST", "non-reserved-(cannot-be-function-or-type)"),
    ("GROUP", "reserved"),
    ("GROUPING", "non-reserved-(cannot-be-function-or-type)"),
    ("GROUPS", "non-reserved"),
    ("HANDLER", "non-reserved"),
    ("HAVING", "reserved"),
    ("HEADER", "non-reserved"),
    ("HEX", "not-keyword"),
    ("HIERARCHY", "not-keyword"),
    ("HOLD", "non-reserved"),
    ("HOUR", "non-reserved"),
    ("ID", "not-keyword"),
    ("IDENTITY", "non-reserved"),
    ("IF", "non-reserved"),
    ("IGNORE", "not-keyword"),
    ("ILIKE", "reserved-(can-be-function-or-type)"),
    ("IMMEDIATE", "non-reserved"),
    ("IMMEDIATELY", "not-keyword"),
    ("IMMUTABLE", "non-reserved"),
    ("IMPLEMENTATION", "not-keyword"),
    ("IMPLICIT", "non-reserved"),
    ("IMPORT", "non-reserved"),
    ("IN", "reserved"),
    ("INCLUDE", "non-reserved"),
    ("INCLUDING", "non-reserved"),
    ("INCREMENT", "non-reserved"),
    ("INDENT", "not-keyword"),
    ("INDEX", "non-reserved"),
    ("INDEXES", "non-reserved"),
    ("INET", "non-reserved-(cannot-be-function-or-type)"),
    ("INDICATOR", "not-keyword"),
    ("INHERIT", "non-reserved"),
    ("INHERITS", "non-reserved"),
    ("INITIAL", "not-keyword"),
    ("INITIALLY", "reserved"),
    ("INLINE", "non-reserved"),
    ("INNER", "reserved-(can-be-function-or-type)"),
    ("INOUT", "non-reserved-(cannot-be-function-or-type)"),
    ("INPUT", "non-reserved"),
    ("INSENSITIVE", "non-reserved"),
    ("INSERT", "non-reserved"),
    ("INSTANCE", "not-keyword"),
    ("INSTANTIABLE", "not-keyword"),
    ("INSTEAD", "non-reserved"),
    ("INT", "non-reserved-(cannot-be-function-or-type)"),
    ("INT2", "non-reserved-(cannot-be-function-or-type)"),
    ("INT4", "non-reserved-(cannot-be-function-or-type)"),
    ("INT4RANGE", "non-reserved-(cannot-be-function-or-type)"),
    ("INT8", "non-reserved-(cannot-be-function-or-type)"),
    ("INT8RANGE", "non-reserved-(cannot-be-function-or-type)"),
    ("INTEGER", "non-reserved-(cannot-be-function-or-type)"),
    ("INTEGRITY", "not-keyword"),
    ("INTERSECT", "reserved"),
    ("INTERSECTION", "not-keyword"),
    ("INTERVAL", "non-reserved-(cannot-be-function-or-type)"),
    ("INTO", "reserved"),
    ("INVOKER", "non-reserved"),
    ("IS", "reserved-(can-be-function-or-type)"),
    ("ISNULL", "reserved-(can-be-function-or-type)"),
    ("ISOLATION", "non-reserved"),
    ("JOIN", "reserved-(can-be-function-or-type)"),
    ("JSON", "not-keyword"),
    ("JSON_ARRAY", "not-keyword"),
    ("JSON_ARRAYAGG", "not-keyword"),
    ("JSON_EXISTS", "not-keyword"),
    ("JSON_OBJECT", "not-keyword"),
    ("JSON_OBJECTAGG", "not-keyword"),
    ("JSON_QUERY", "not-keyword"),
    ("JSON_TABLE", "not-keyword"),
    ("JSON_TABLE_PRIMITIVE", "not-keyword"),
    ("JSON_VALUE", "not-keyword"),
    ("JSONB", "non-reserved-(cannot-be-function-or-type)"),
    ("K", "not-keyword"),
    ("KEEP", "not-keyword"),
    ("KEY", "non-reserved"),
    ("KEYS", "not-keyword"),
    ("KEY_MEMBER", "not-keyword"),
    ("KEY_TYPE", "not-keyword"),
    ("LABEL", "non-reserved"),
    ("LAG", "not-keyword"),
    ("LANGUAGE", "non-reserved"),
    ("LARGE", "non-reserved"),
    ("LAST", "non-reserved"),
    ("LAST_VALUE", "not-keyword"),
    ("LATERAL", "reserved"),
    ("LEAD", "not-keyword"),
    ("LEADING", "reserved"),
    ("LEAKPROOF", "non-reserved"),
    ("LEAST", "non-reserved-(cannot-be-function-or-type)"),
    ("LEFT", "reserved-(can-be-function-or-type)"),
    ("LENGTH", "not-keyword"),
    ("LEVEL", "non-reserved"),
    ("LIBRARY", "not-keyword"),
    ("LIKE", "reserved-(can-be-function-or-type)"),
    ("LIKE_REGEX", "not-keyword"),
    ("LIMIT", "reserved"),
    ("LINE", "non-reserved-(cannot-be-function-or-type)"),
    ("LINK", "not-keyword"),
    ("LISTAGG", "not-keyword"),
    ("LISTEN", "non-reserved"),
    ("LN", "not-keyword"),
    ("LOAD", "non-reserved"),
    ("LOCAL", "non-reserved"),
    ("LOCALTIME", "reserved"),
    ("LOCALTIMESTAMP", "reserved"),
    ("LOCATION", "non-reserved"),
    ("LOCATOR", "not-keyword"),
    ("LOCK", "non-reserved"),
    ("LOCKED", "non-reserved"),
    ("LOG", "not-keyword"),
    ("LOG10", "not-keyword"),
    ("LOGGED", "non-reserved"),
    ("LOWER", "not-keyword"),
    ("LSEG", "non-reserved-(cannot-be-function-or-type)"),
    ("M", "not-keyword"),
    ("MACADDR", "non-reserved-(cannot-be-function-or-type)"),
    ("MACADDR8", "non-reserved-(cannot-be-function-or-type)"),
    ("MAP", "not-keyword"),
    ("MAPPING", "non-reserved"),
    ("MATCH", "non-reserved"),
    ("MATCHED", "non-reserved"),
    ("MATCHES", "not-keyword"),
    ("MATCH_NUMBER", "not-keyword"),
    ("MATCH_RECOGNIZE", "not-keyword"),
    ("MATERIALIZED", "non-reserved"),
    ("MAX", "not-keyword"),
    ("MAXVALUE", "non-reserved"),
    ("MEASURES", "not-keyword"),
    ("MEMBER", "not-keyword"),
    ("MERGE", "non-reserved"),
    ("MESSAGE_LENGTH", "not-keyword"),
    ("MESSAGE_OCTET_LENGTH", "not-keyword"),
    ("MESSAGE_TEXT", "not-keyword"),
    ("METHOD", "non-reserved"),
    ("MIN", "not-keyword"),
    ("MINUTE", "non-reserved"),
    ("MINVALUE", "non-reserved"),
    ("MOD", "not-keyword"),
    ("MODE", "non-reserved"),
    ("MODIFIES", "not-keyword"),
    ("MODULE", "not-keyword"),
    ("MONEY", "non-reserved-(cannot-be-function-or-type)"),
    ("MONTH", "non-reserved"),
    ("MORE", "not-keyword"),
    ("MOVE", "non-reserved"),
    ("MULTISET", "not-keyword"),
    ("MUMPS", "not-keyword"),
    ("NAME", "non-reserved"),
    ("NAMES", "non-reserved"),
    ("NAMESPACE", "not-keyword"),
    ("NATIONAL", "non-reserved-(cannot-be-function-or-type)"),
    ("NATURAL", "reserved-(can-be-function-or-type)"),
    ("NCHAR", "non-reserved-(cannot-be-function-or-type)"),
    ("NCLOB", "not-keyword"),
    ("NESTED", "not-keyword"),
    ("NESTING", "not-keyword"),
    ("NEW", "non-reserved"),
    ("NEXT", "non-reserved"),
    ("NFC", "non-reserved"),
    ("NFD", "non-reserved"),
    ("NFKC", "non-reserved"),
    ("NFKD", "non-reserved"),
    ("NIL", "not-keyword"),
    ("NO", "non-reserved"),
    ("NONE", "non-reserved-(cannot-be-function-or-type)"),
    ("NORMALIZE", "non-reserved-(cannot-be-function-or-type)"),
    ("NORMALIZED", "non-reserved"),
    ("NOT", "reserved"),
    ("NOTHING", "non-reserved"),
    ("NOTIFY", "non-reserved"),
    ("NOTNULL", "reserved-(can-be-function-or-type)"),
    ("NOWAIT", "non-reserved"),
    ("NTH_VALUE", "not-keyword"),
    ("NTILE", "not-keyword"),
    ("NULL", "reserved"),
    ("NULLABLE", "not-keyword"),
    ("NULLIF", "non-reserved-(cannot-be-function-or-type)"),
    ("NULLS", "non-reserved"),
    ("NUMBER", "not-keyword"),
    ("NUMERIC", "non-reserved-(cannot-be-function-or-type)"),
    ("NUMRANGE", "non-reserved-(cannot-be-function-or-type)"),
    ("OBJECT", "non-reserved"),
    ("OCCURRENCES_REGEX", "not-keyword"),
    ("OCTETS", "not-keyword"),
    ("OCTET_LENGTH", "not-keyword"),
    ("OF", "non-reserved"),
    ("OFF", "non-reserved"),
    ("OFFSET", "reserved"),
    ("OIDS", "non-reserved"),
    ("OLD", "non-reserved"),
    ("OMIT", "not-keyword"),
    ("ON", "reserved"),
    ("ONE", "not-keyword"),
    ("ONLY", "reserved"),
    ("OPEN", "not-keyword"),
    ("OPERATOR", "non-reserved"),
    ("OPTION", "non-reserved"),
    ("OPTIONS", "non-reserved"),
    ("OR", "reserved"),
    ("ORDER", "reserved"),
    ("ORDERING", "not-keyword"),
    ("ORDINALITY", "non-reserved"),
    ("OTHERS", "non-reserved"),
    ("OUT", "non-reserved-(cannot-be-function-or-type)"),
    ("OUTER", "reserved-(can-be-function-or-type)"),
    ("OUTPUT", "not-keyword"),
    ("OVER", "non-reserved"),
    ("OVERFLOW", "not-keyword"),
    ("OVERLAPS", "reserved-(can-be-function-or-type)"),
    ("OVERLAY", "non-reserved-(cannot-be-function-or-type)"),
    ("OVERRIDING", "non-reserved"),
    ("OWNED", "non-reserved"),
    ("OWNER", "non-reserved"),
    ("P", "not-keyword"),
    ("PAD", "not-keyword"),
    ("PARALLEL", "non-reserved"),
    ("PARAMETER", "not-keyword"),
    ("PARAMETER_MODE", "not-keyword"),
    ("PARAMETER_NAME", "not-keyword"),
    ("PARAMETER_ORDINAL_POSITION", "not-keyword"),
    ("PARAMETER_SPECIFIC_CATALOG", "not-keyword"),
    ("PARAMETER_SPECIFIC_NAME", "not-keyword"),
    ("PARAMETER_SPECIFIC_SCHEMA", "not-keyword"),
    ("PARSER", "non-reserved"),
    ("PARTIAL", "non-reserved"),
    ("PARTITION", "non-reserved"),
    ("PASCAL", "not-keyword"),
    ("PASS", "not-keyword"),
    ("PASSING", "non-reserved"),
    ("PASSTHROUGH", "not-keyword"),
    ("PASSWORD", "non-reserved"),
    ("PAST", "not-keyword"),
    ("PATH", "non-reserved-(cannot-be-function-or-type)"),
    ("PATTERN", "not-keyword"),
    ("PER", "not-keyword"),
    ("PERCENT", "not-keyword"),
    ("PERCENTILE_CONT", "not-keyword"),
    ("PERCENTILE_DISC", "not-keyword"),
    ("PERCENT_RANK", "not-keyword"),
    ("PERIOD", "not-keyword"),
    ("PERMISSION", "not-keyword"),
    ("PERMISSIVE", "non-reserved"),
    ("PERMUTE", "not-keyword"),
    ("PG_LSN", "non-reserved-(cannot-be-function-or-type)"),
    ("PLACING", "reserved"),
    ("PLAN", "not-keyword"),
    ("PLANS", "non-reserved"),
    ("PLI", "not-keyword"),
    ("POINT", "non-reserved-(cannot-be-function-or-type)"),
    ("POLICY", "non-reserved"),
    ("POLYGON", "non-reserved-(cannot-be-function-or-type)"),
    ("PORTION", "not-keyword"),
    ("POSITION", "non-reserved-(cannot-be-function-or-type)"),
    ("POSITION_REGEX", "not-keyword"),
    ("POWER", "not-keyword"),
    ("PRECEDES", "not-keyword"),
    ("PRECEDING", "non-reserved"),
    ("PRECISION", "non-reserved-(cannot-be-function-or-type)"),
    ("PREPARE", "non-reserved"),
    ("PREPARED", "non-reserved"),
    ("PRESERVE", "non-reserved"),
    ("PRIMARY", "reserved"),
    ("PRIOR", "non-reserved"),
    ("PRIVATE", "not-keyword"),
    ("PRIVILEGES", "non-reserved"),
    ("PROCEDURAL", "non-reserved"),
    ("PROCEDURE", "non-reserved"),
    ("PROCEDURES", "non-reserved"),
    ("PROGRAM", "non-reserved"),
    ("PRUNE", "not-keyword"),
    ("PTF", "not-keyword"),
    ("PUBLIC", "not-keyword"),
    ("PUBLICATION", "non-reserved"),
    ("QUOTE", "non-reserved"),
    ("QUOTES", "not-keyword"),
    ("RANGE", "non-reserved"),
    ("RANK", "not-keyword"),
    ("READ", "non-reserved"),
    ("READS", "not-keyword"),
    ("REAL", "non-reserved-(cannot-be-function-or-type)"),
    ("REASSIGN", "non-reserved"),
    ("RECHECK", "non-reserved"),
    ("RECOVERY", "not-keyword"),
    ("RECURSIVE", "non-reserved"),
    ("REF", "non-reserved"),
    ("REFERENCES", "reserved"),
    ("REFERENCING", "non-reserved"),
    ("REFRESH", "non-reserved"),
    ("REGR_AVGX", "not-keyword"),
    ("REGR_AVGY", "not-keyword"),
    ("REGR_COUNT", "not-keyword"),
    ("REGR_INTERCEPT", "not-keyword"),
    ("REGR_R2", "not-keyword"),
    ("REGR_SLOPE", "not-keyword"),
    ("REGR_SXX", "not-keyword"),
    ("REGR_SXY", "not-keyword"),
    ("REGR_SYY", "not-keyword"),
    ("REINDEX", "non-reserved"),
    ("RELATIVE", "non-reserved"),
    ("RELEASE", "non-reserved"),
    ("RENAME", "non-reserved"),
    ("REPEATABLE", "non-reserved"),
    ("REPLACE", "non-reserved"),
    ("REPLICA", "non-reserved"),
    ("REQUIRING", "not-keyword"),
    ("RESET", "non-reserved"),
    ("RESPECT", "not-keyword"),
    ("RESTART", "non-reserved"),
    ("RESTORE", "not-keyword"),
    ("RESTRICT", "non-reserved"),
    ("RESTRICTIVE", "non-reserved"),
    ("RESULT", "not-keyword"),
    ("RETURN", "not-keyword"),
    ("RETURNED_CARDINALITY", "not-keyword"),
    ("RETURNED_LENGTH", "not-keyword"),
    ("RETURNED_OCTET_LENGTH", "not-keyword"),
    ("RETURNED_SQLSTATE", "not-keyword"),
    ("RETURNING", "reserved"),
    ("RETURNS", "non-reserved"),
    ("REVOKE", "non-reserved"),
    ("RIGHT", "reserved-(can-be-function-or-type)"),
    ("ROLE", "non-reserved"),
    ("ROLLBACK", "non-reserved"),
    ("ROLLUP", "non-reserved"),
    ("ROUTINE", "non-reserved"),
    ("ROUTINES", "non-reserved"),
    ("ROUTINE_CATALOG", "not-keyword"),
    ("ROUTINE_NAME", "not-keyword"),
    ("ROUTINE_SCHEMA", "not-keyword"),
    ("ROW", "non-reserved-(cannot-be-function-or-type)"),
    ("ROWS", "non-reserved"),
    ("ROW_COUNT", "not-keyword"),
    ("ROW_NUMBER", "not-keyword"),
    ("RULE", "non-reserved"),
    ("RUNNING", "not-keyword"),
    ("SAVEPOINT", "non-reserved"),
    ("SCALAR", "not-keyword"),
    ("SCALE", "not-keyword"),
    ("SCHEMA", "non-reserved"),
    ("SCHEMAS", "non-reserved"),
    ("SCHEMA_NAME", "not-keyword"),
    ("SCOPE", "not-keyword"),
    ("SCOPE_CATALOG", "not-keyword"),
    ("SCOPE_NAME", "not-keyword"),
    ("SCOPE_SCHEMA", "not-keyword"),
    ("SCROLL", "non-reserved"),
    ("SEARCH", "non-reserved"),
    ("SECOND", "non-reserved"),
    ("SECTION", "not-keyword"),
    ("SECURITY", "non-reserved"),
    ("SEEK", "not-keyword"),
    ("SELECT", "reserved"),
    ("SELECTIVE", "not-keyword"),
    ("SELF", "not-keyword"),
    ("SENSITIVE", "not-keyword"),
    ("SEQUENCE", "non-reserved"),
    ("SEQUENCES", "non-reserved"),
    ("SERIAL", "non-reserved-(cannot-be-function-or-type)"),
    ("SERIAL2", "non-reserved-(cannot-be-function-or-type)"),
    ("SERIAL4", "non-reserved-(cannot-be-function-or-type)"),
    ("SERIAL8", "non-reserved-(cannot-be-function-or-type)"),
    ("SERIALIZABLE", "non-reserved"),
    ("SERVER", "non-reserved"),
    ("SERVER_NAME", "not-keyword"),
    ("SESSION", "non-reserved"),
    ("SESSION_USER", "reserved"),
    ("SET", "non-reserved"),
    ("SETOF", "non-reserved-(cannot-be-function-or-type)"),
    ("SETS", "non-reserved"),
    ("SHARE", "non-reserved"),
    ("SHOW", "non-reserved"),
    ("SIMILAR", "reserved-(can-be-function-or-type)"),
    ("SIMPLE", "non-reserved"),
    ("SIN", "not-keyword"),
    ("SINH", "not-keyword"),
    ("SIZE", "not-keyword"),
    ("SKIP", "non-reserved"),
    ("SMALLINT", "non-reserved-(cannot-be-function-or-type)"),
    ("SMALLSERIAL", "non-reserved-(cannot-be-function-or-type)"),
    ("SNAPSHOT", "non-reserved"),
    ("SOME", "reserved"),
    ("SOURCE", "not-keyword"),
    ("SPACE", "not-keyword"),
    ("SPECIFIC", "not-keyword"),
    ("SPECIFICTYPE", "not-keyword"),
    ("SPECIFIC_NAME", "not-keyword"),
    ("SQL", "non-reserved"),
    ("SQLCODE", "not-keyword"),
    ("SQLERROR", "not-keyword"),
    ("SQLEXCEPTION", "not-keyword"),
    ("SQLSTATE", "not-keyword"),
    ("SQLWARNING", "not-keyword"),
    ("SQRT", "not-keyword"),
    ("STABLE", "non-reserved"),
    ("STANDALONE", "non-reserved"),
    ("START", "non-reserved"),
    ("STATE", "not-keyword"),
    ("STATEMENT", "non-reserved"),
    ("STATIC", "not-keyword"),
    ("STATISTICS", "non-reserved"),
    ("STDDEV_POP", "not-keyword"),
    ("STDDEV_SAMP", "not-keyword"),
    ("STDIN", "non-reserved"),
    ("STDOUT", "non-reserved"),
    ("STORAGE", "non-reserved"),
    ("STORED", "non-reserved"),
    ("STRICT", "non-reserved"),
    ("STRING", "not-keyword"),
    ("STRIP", "non-reserved"),
    ("STRUCTURE", "not-keyword"),
    ("STYLE", "not-keyword"),
    ("SUBCLASS_ORIGIN", "not-keyword"),
    ("SUBMULTISET", "not-keyword"),
    ("SUBSCRIPTION", "non-reserved"),
    ("SUBSET", "not-keyword"),
    ("SUBSTRING", "non-reserved-(cannot-be-function-or-type)"),
    ("SUBSTRING_REGEX", "not-keyword"),
    ("SUCCEEDS", "not-keyword"),
    ("SUM", "not-keyword"),
    ("SUPPORT", "non-reserved"),
    ("SYMMETRIC", "reserved"),
    ("SYSID", "non-reserved"),
    ("SYSTEM", "non-reserved"),
    ("SYSTEM_TIME", "not-keyword"),
    ("SYSTEM_USER", "not-keyword"),
    ("T", "not-keyword"),
    ("TABLE", "non-reserved"),
    ("TABLES", "non-reserved"),
    ("TABLESAMPLE", "reserved-(can-be-function-or-type)"),
    ("TABLESPACE", "non-reserved"),
    ("TABLE_NAME", "not-keyword"),
    ("TAN", "not-keyword"),
    ("TANH", "not-keyword"),
    ("TEMP", "non-reserved"),
    ("TEMPLATE", "non-reserved"),
    ("TEMPORARY", "non-reserved"),
    ("TEXT", "non-reserved"),
    ("THEN", "reserved"),
    ("THROUGH", "not-keyword"),
    ("TIES", "non-reserved"),
    ("TIME", "non-reserved-(cannot-be-function-or-type)"),
    ("TIMESTAMP", "non-reserved-(cannot-be-function-or-type)"),
    ("TIMEZONE_HOUR", "not-keyword"),
    ("TIMEZONE_MINUTE", "not-keyword"),
    ("TO", "reserved"),
    ("TOKEN", "not-keyword"),
    ("TOP_LEVEL_COUNT", "not-keyword"),
    ("TRAILING", "reserved"),
    ("TRANSACTION", "non-reserved"),
    ("TRANSACTIONS_COMMITTED", "not-keyword"),
    ("TRANSACTIONS_ROLLED_BACK", "not-keyword"),
    ("TRANSACTION_ACTIVE", "not-keyword"),
    ("TRANSFORM", "non-reserved"),
    ("TRANSFORMS", "not-keyword"),
    ("TRANSLATE", "not-keyword"),
    ("TRANSLATE_REGEX", "not-keyword"),
    ("TRANSLATION", "not-keyword"),
    ("TREAT", "non-reserved-(cannot-be-function-or-type)"),
    ("TRIGGER", "non-reserved"),
    ("TRIGGER_CATALOG", "not-keyword"),
    ("TRIGGER_NAME", "not-keyword"),
    ("TRIGGER_SCHEMA", "not-keyword"),
    ("TRIM", "non-reserved-(cannot-be-function-or-type)"),
    ("TRIM_ARRAY", "not-keyword"),
    ("TRUE", "reserved"),
    ("TRUNCATE", "non-reserved"),
    ("TRUSTED", "non-reserved"),
    ("TSQUERY", "non-reserved-(cannot-be-function-or-type)"),
    ("TSRANGE", "non-reserved-(cannot-be-function-or-type)"),
    ("TSTZRANGE", "non-reserved-(cannot-be-function-or-type)"),
    ("TSVECTOR", "non-reserved-(cannot-be-function-or-type)"),
    ("TYPE", "non-reserved"),
    ("TYPES", "non-reserved"),
    ("UESCAPE", "non-reserved"),
    ("UNBOUNDED", "non-reserved"),
    ("UNCOMMITTED", "non-reserved"),
    ("UNCONDITIONAL", "not-keyword"),
    ("UNDER", "not-keyword"),
    ("UNENCRYPTED", "non-reserved"),
    ("UNION", "reserved"),
    ("UNIQUE", "reserved"),
    ("UNKNOWN", "non-reserved"),
    ("UNLINK", "not-keyword"),
    ("UNLISTEN", "non-reserved"),
    ("UNLOGGED", "non-reserved"),
    ("UNMATCHED", "not-keyword"),
    ("UNNAMED", "not-keyword"),
    ("UNNEST", "not-keyword"),
    ("UNTIL", "non-reserved"),
    ("UNTYPED", "not-keyword"),
    ("UPDATE", "non-reserved"),
    ("UPPER", "not-keyword"),
    ("URI", "not-keyword"),
    ("USAGE", "not-keyword"),
    ("USER", "reserved"),
    ("USER_DEFINED_TYPE_CATALOG", "not-keyword"),
    ("USER_DEFINED_TYPE_CODE", "not-keyword"),
    ("USER_DEFINED_TYPE_NAME", "not-keyword"),
    ("USER_DEFINED_TYPE_SCHEMA", "not-keyword"),
    ("USING", "reserved"),
    ("UTF16", "not-keyword"),
    ("UTF32", "not-keyword"),
    ("UTF8", "not-keyword"),
    ("UUID", "non-reserved-(cannot-be-function-or-type)"),
    ("VACUUM", "non-reserved"),
    ("VALID", "non-reserved"),
    ("VALIDATE", "non-reserved"),
    ("VALIDATOR", "non-reserved"),
    ("VALUE", "non-reserved"),
    ("VALUES", "non-reserved-(cannot-be-function-or-type)"),
    ("VALUE_OF", "not-keyword"),
    ("VARBINARY", "not-keyword"),
    ("VARCHAR", "non-reserved-(cannot-be-function-or-type)"),
    ("VARIADIC", "reserved"),
    ("VARYING", "non-reserved"),
    ("VAR_POP", "not-keyword"),
    ("VAR_SAMP", "not-keyword"),
    ("VERBOSE", "reserved-(can-be-function-or-type)"),
    ("VERSION", "non-reserved"),
    ("VERSIONING", "not-keyword"),
    ("VIEW", "non-reserved"),
    ("VIEWS", "non-reserved"),
    ("VOLATILE", "non-reserved"),
    ("WHEN", "reserved"),
    ("WHENEVER", "not-keyword"),
    ("WHERE", "reserved"),
    ("WHITESPACE", "non-reserved"),
    ("WIDTH_BUCKET", "not-keyword"),
    ("WINDOW", "reserved"),
    ("WITH", "reserved"),
    ("WITHIN", "non-reserved"),
    ("WITHOUT", "non-reserved"),
    ("WORK", "non-reserved"),
    ("WRAPPER", "non-reserved"),
    ("WRITE", "non-reserved"),
    ("XML", "non-reserved"),
    ("XMLAGG", "not-keyword"),
    ("XMLATTRIBUTES", "non-reserved-(cannot-be-function-or-type)"),
    ("XMLBINARY", "not-keyword"),
    ("XMLCAST", "not-keyword"),
    ("XMLCOMMENT", "not-keyword"),
    ("XMLCONCAT", "non-reserved-(cannot-be-function-or-type)"),
    ("XMLDECLARATION", "not-keyword"),
    ("XMLDOCUMENT", "not-keyword"),
    ("XMLELEMENT", "non-reserved-(cannot-be-function-or-type)"),
    ("XMLEXISTS", "non-reserved-(cannot-be-function-or-type)"),
    ("XMLFOREST", "non-reserved-(cannot-be-function-or-type)"),
    ("XMLITERATE", "not-keyword"),
    ("XMLNAMESPACES", "non-reserved-(cannot-be-function-or-type)"),
    ("XMLPARSE", "non-reserved-(cannot-be-function-or-type)"),
    ("XMLPI", "non-reserved-(cannot-be-function-or-type)"),
    ("XMLQUERY", "not-keyword"),
    ("XMLROOT", "non-reserved-(cannot-be-function-or-type)"),
    ("XMLSCHEMA", "not-keyword"),
    ("XMLSERIALIZE", "non-reserved-(cannot-be-function-or-type)"),
    ("XMLTABLE", "non-reserved-(cannot-be-function-or-type)"),
    ("XMLTEXT", "not-keyword"),
    ("XMLVALIDATE", "not-keyword"),
    ("YEAR", "non-reserved"),
    ("YES", "non-reserved"),
    ("ZONE", "non-reserved"),
]

postgres_nondocs_keywords = [
    ("ALLOW_CONNECTIONS", "non-reserved"),
    ("BREADTH", "non-reserved"),
    ("BUFFERS", "non-reserved"),
    ("BYPASSRLS", "non-reserved"),
    ("CONNECT", "reserved"),
    ("COSTS", "non-reserved"),
    ("CURRENT_USER", "non-reserved"),
    ("CREATEDB", "non-reserved"),
    ("CREATEROLE", "non-reserved"),
    ("DATE", "non-reserved"),
    ("DEPTH", "non-reserved"),
    ("DESCRIBE", "non-reserved"),
    ("DETERMINISTIC", "non-reserved"),
    ("EXTENDED", "non-reserved"),
    ("FILE", "non-reserved"),
    ("FORCE_NOT_NULL", "non-reserved"),
    ("FORCE_NULL", "non-reserved"),
    ("FORCE_QUOTE", "non-reserved"),
    ("FORMAT", "non-reserved"),
    ("HASH", "non-reserved"),
    ("ICU", "non-reserved"),
    ("IGNORE", "non-reserved"),
    ("IS_TEMPLATE", "non-reserved"),
    ("JSON", "non-reserved"),
    ("LC_COLLATE", "non-reserved"),
    ("LC_CTYPE", "non-reserved"),
    ("LIBC", "non-reserved"),
    ("LIST", "non-reserved"),
    ("LOGIN", "non-reserved"),
    ("LOCALE", "non-reserved"),
    ("MAIN", "non-reserved"),
    ("MODULUS", "non-reserved"),
    ("NOBYPASSRLS", "non-reserved"),
    ("NOCREATEDB", "non-reserved"),
    ("NOCREATEROLE", "non-reserved"),
    ("NOINHERIT", "non-reserved"),
    ("NOLOGIN", "non-reserved"),
    ("NOREPLICATION", "non-reserved"),
    ("NOSUPERUSER", "non-reserved"),
    ("PLAIN", "non-reserved"),
    ("PROVIDER", "non-reserved"),
    ("PUBLIC", "non-reserved"),
    ("REMAINDER", "non-reserved"),
    ("REPLICATION", "non-reserved"),
    ("RESPECT", "non-reserved"),
    ("RESTRICTED", "non-reserved"),
    ("SAFE", "non-reserved"),
    ("SETTINGS", "non-reserved"),
    ("SKIP_LOCKED", "non-reserved"),
    ("SUMMARY", "non-reserved"),
    ("SUPERUSER", "non-reserverd"),
    ("TIMETZ", "non-reserved"),
    ("TIMESTAMPTZ", "non-reserved"),
    ("TIMING", "non-reserved"),
    ("UNSAFE", "non-reserved"),
    ("USAGE", "non-reserved"),
    ("WAL", "non-reserved"),
]

postgres_postgis_datatype_keywords = [
    ("POINT", "non-reserved"),
    ("LINESTRING", "non-reserved"),
    ("POLYGON", "non-reserved"),
    ("MULTIPOINT", "non-reserved"),
    ("MULTILINESTRING", "non-reserved"),
    ("MULTIPOLYGON", "non-reserved"),
    ("GEOMETRYCOLLECTION", "non-reserved"),
    ("POINTZ", "non-reserved"),
    ("LINESTRINGZ", "non-reserved"),
    ("POLYGONZ", "non-reserved"),
    ("MULTIPOINTZ", "non-reserved"),
    ("MULTILINESTRINGZ", "non-reserved"),
    ("MULTIPOLYGONZ", "non-reserved"),
    ("GEOMETRYCOLLECTIONZ", "non-reserved"),
    ("POINTM", "non-reserved"),
    ("LINESTRINGM", "non-reserved"),
    ("POLYGONM", "non-reserved"),
    ("MULTIPOINTM", "non-reserved"),
    ("MULTILINESTRINGM", "non-reserved"),
    ("MULTIPOLYGONM", "non-reserved"),
    ("GEOMETRYCOLLECTIONM", "non-reserved"),
    ("POINTZM", "non-reserved"),
    ("LINESTRINGZM", "non-reserved"),
    ("POLYGONZM", "non-reserved"),
    ("MULTIPOINTZM", "non-reserved"),
    ("MULTILINESTRINGZM", "non-reserved"),
    ("MULTIPOLYGONZM", "non-reserved"),
    ("GEOMETRYCOLLECTIONZM", "non-reserved"),
    ("CIRCULARSTRING", "non-reserved"),
    ("COMPOUNDCURVE", "non-reserved"),
    ("CURVEPOLYGON", "non-reserved"),
    ("MULTICURVE", "non-reserved"),
    ("MULTISURFACE", "non-reserved"),
    ("POLYHEDRALSURFACE", "non-reserved"),
    ("TRIANGLE", "non-reserved"),
    ("TIN", "non-reserved"),
]

postgres_postgis_other_keywords = [
    ("GEOMETRY", "non-reserved"),
    ("GEOGRAPHY", "non-reserved"),
    ("EMPTY", "non-reserved"),
]

postgres_keywords = priority_keyword_merge(
    postgres_docs_keywords,
    postgres_nondocs_keywords,
    postgres_postgis_datatype_keywords,
    postgres_postgis_other_keywords,
)



"""Tests the python routines within L027."""

import sqlfluff


def test__rules__std_L027_wildcard_single_count():
    """Verify that L027 is only raised once for wildcard (see issue #1973)."""
    sql = """
        SELECT *
        FROM foo
        INNER JOIN bar;
    """
    result = sqlfluff.lint(sql)
    assert "L027" in [r["code"] for r in result]
    assert [r["code"] for r in result].count("L027") == 1



"""Tests the python routines within L020."""

import sqlfluff


def test__rules__std_L020_one_aliases_one_duplicate():
    """Verify correct error message for one duplicate table aliases occur one times."""
    sql = """
        SELECT
            a.pk
        FROM table_1 AS a
        JOIN table_2 AS a ON a.pk = a.pk
    """
    result = sqlfluff.lint(sql)
    assert "L020" in [r["code"] for r in result]
    assert [r["code"] for r in result].count("L020") == 1


def test__rules__std_L020_one_aliases_two_duplicate():
    """Verify correct error message for one duplicate table aliases occur two times."""
    sql = """
        SELECT
            a.pk
        FROM table_1 AS a
        JOIN table_2 AS a ON a.pk = a.pk
        JOIN table_3 AS a ON a.pk = a.pk
    """
    result = sqlfluff.lint(sql)
    result_filter = [r for r in result if r["code"] == "L020"]
    # Error message only show two times, not three
    assert len(result_filter) == 2
    assert (
        len(
            [
                r
                for r in result_filter
                if "Duplicate table alias 'a'" in r["description"]
            ]
        )
        == 2
    )
    # Test specific line number
    assert result_filter[0]["line_no"] == 5
    assert result_filter[1]["line_no"] == 6


def test__rules__std_L020_complex():
    """Verify that L020 returns the correct error message for complex example."""
    sql = """
        SELECT
            a.pk,
            b.pk
        FROM table_1 AS a
        JOIN table_2 AS a ON a.pk = a.pk
        JOIN table_3 AS b ON a.pk = b.pk
        JOIN table_4 AS b ON b.pk = b.pk
        JOIN table_5 AS a ON b.pk = a.pk
    """
    result = sqlfluff.lint(sql)
    result_filter = [r for r in result if r["code"] == "L020"]
    # Error message only show two times, not three
    assert len(result_filter) == 3
    assert (
        len(
            [
                r
                for r in result_filter
                if "Duplicate table alias 'a'" in r["description"]
            ]
        )
        == 2
    )
    assert (
        len(
            [
                r
                for r in result_filter
                if "Duplicate table alias 'b'" in r["description"]
            ]
        )
        == 1
    )
    # Test specific line number
    assert result_filter[0]["line_no"] == 6
    assert result_filter[1]["line_no"] == 8
    assert result_filter[2]["line_no"] == 9



"""'Trace' Jinja template execution to map output back to the raw template.

This is a newer slicing algorithm that handles cases heuristic.py does not.
"""

from dataclasses import dataclass, field
import logging
import regex
from typing import Callable, cast, Dict, List, NamedTuple, Optional, Tuple

from jinja2 import Environment
from jinja2.environment import Template
from jinja2.exceptions import TemplateSyntaxError

from sqlfluff.core.templaters.base import (
    RawFileSlice,
    TemplatedFileSlice,
)


# Instantiate the templater logger
templater_logger = logging.getLogger("sqlfluff.templater")


class JinjaTrace(NamedTuple):
    """Returned by JinjaTracer.trace()."""

    # Template output
    templated_str: str
    # Raw (i.e. before rendering) Jinja template sliced into tokens
    raw_sliced: List[RawFileSlice]
    # Rendered Jinja template (i.e. output) mapped back to rwa_str source
    sliced_file: List[TemplatedFileSlice]


@dataclass
class RawSliceInfo:
    """JinjaTracer-specific info about each RawFileSlice."""

    unique_alternate_id: Optional[str]
    alternate_code: Optional[str]
    next_slice_indices: List[int] = field(default_factory=list)
    inside_block: bool = field(default=False)  # {% block %}


class JinjaTracer:
    """Records execution path of a Jinja template."""

    def __init__(
        self,
        raw_str: str,
        raw_sliced: List[RawFileSlice],
        raw_slice_info: Dict[RawFileSlice, RawSliceInfo],
        sliced_file: List[TemplatedFileSlice],
        make_template: Callable[[str], Template],
    ):
        # Input
        self.raw_str = raw_str
        self.raw_sliced = raw_sliced
        self.raw_slice_info = raw_slice_info
        self.sliced_file = sliced_file
        self.make_template = make_template

        # Internal bookkeeping
        self.program_counter: int = 0
        self.source_idx: int = 0

    def trace(self, append_to_templated: str = "") -> JinjaTrace:
        """Executes raw_str. Returns template output and trace."""
        trace_template_str = "".join(
            cast(str, self.raw_slice_info[rs].alternate_code)
            if self.raw_slice_info[rs].alternate_code is not None
            else rs.raw
            for rs in self.raw_sliced
        )
        trace_template = self.make_template(trace_template_str)
        trace_template_output = trace_template.render()
        # Split output by section. Each section has two possible formats.
        trace_entries = list(regex.finditer(r"\0", trace_template_output))
        for match_idx, match in enumerate(trace_entries):
            pos1 = match.span()[0]
            try:
                pos2 = trace_entries[match_idx + 1].span()[0]
            except IndexError:
                pos2 = len(trace_template_output)
            p = trace_template_output[pos1 + 1 : pos2]
            m_id = regex.match(r"^([0-9a-f]+)(_(\d+))?", p)
            if not m_id:
                raise ValueError(  # pragma: no cover
                    "Internal error. Trace template output does not match expected "
                    "format."
                )
            if m_id.group(3):
                # E.g. "00000000000000000000000000000001_83". The number after
                # "_" is the length (in characters) of a corresponding literal
                # in raw_str.
                value = [m_id.group(1), int(m_id.group(3)), True]
            else:
                # E.g. "00000000000000000000000000000002 a < 10". The characters
                # after the slice ID are executable code from raw_str.
                value = [m_id.group(0), p[len(m_id.group(0)) + 1 :], False]
            alt_id, content_info, literal = value
            target_slice_idx = self.find_slice_index(alt_id)
            slice_length = content_info if literal else len(str(content_info))
            target_inside_block = self.raw_slice_info[
                self.raw_sliced[target_slice_idx]
            ].inside_block
            if not target_inside_block:
                # Normal case: Walk through the template.
                self.move_to_slice(target_slice_idx, slice_length)
            else:
                # {% block %} executes code elsewhere in the template but does
                # not move there. It's a bit like macro invocation.
                self.record_trace(slice_length, target_slice_idx)

        # TRICKY: The 'append_to_templated' parameter is only used by the dbt
        # templater, passing "\n" for this parameter if we need to add one back.
        # (The Jinja templater does not pass this parameter, so
        # 'append_to_templated' gets the default value of "", empty string.)
        # For more detail, see the comments near the call to slice_file() in
        # plugins/sqlfluff-templater-dbt/sqlfluff_templater_dbt/templater.py.
        templated_str = self.make_template(self.raw_str).render() + append_to_templated
        return JinjaTrace(templated_str, self.raw_sliced, self.sliced_file)

    def find_slice_index(self, slice_identifier) -> int:
        """Given a slice identifier, return its index.

        A slice identifier is a string like 00000000000000000000000000000002.
        """
        raw_slices_search_result = [
            idx
            for idx, rs in enumerate(self.raw_sliced)
            if self.raw_slice_info[rs].unique_alternate_id == slice_identifier
        ]
        if len(raw_slices_search_result) != 1:
            raise ValueError(  # pragma: no cover
                f"Internal error. Unable to locate slice for {slice_identifier}."
            )
        return raw_slices_search_result[0]

    def move_to_slice(self, target_slice_idx, target_slice_length):
        """Given a template location, walk execution to that point."""
        while self.program_counter < len(self.raw_sliced):
            self.record_trace(
                target_slice_length if self.program_counter == target_slice_idx else 0
            )
            current_raw_slice = self.raw_sliced[self.program_counter]
            if self.program_counter == target_slice_idx:
                # Reached the target slice. Go to next location and stop.
                self.program_counter += 1
                break
            else:
                # Choose the next step.

                # We could simply go to the next slice (sequential execution).
                candidates = [self.program_counter + 1]
                # If we have other options, consider those.
                for next_slice_idx in self.raw_slice_info[
                    current_raw_slice
                ].next_slice_indices:
                    # It's a valid possibility if it does not take us past the
                    # target.
                    if next_slice_idx <= target_slice_idx:
                        candidates.append(next_slice_idx)
                # Choose the candidate that takes us closest to the target.
                candidates.sort(key=lambda c: abs(target_slice_idx - c))
                self.program_counter = candidates[0]

    def record_trace(self, target_slice_length, slice_idx=None, slice_type=None):
        """Add the specified (default: current) location to the trace."""
        if slice_idx is None:
            slice_idx = self.program_counter
        if slice_type is None:
            slice_type = self.raw_sliced[slice_idx].slice_type
        self.sliced_file.append(
            TemplatedFileSlice(
                slice_type,
                slice(
                    self.raw_sliced[slice_idx].source_idx,
                    self.raw_sliced[slice_idx + 1].source_idx
                    if slice_idx + 1 < len(self.raw_sliced)
                    else len(self.raw_str),
                ),
                slice(self.source_idx, self.source_idx + target_slice_length),
            )
        )
        if slice_type in ("literal", "templated"):
            self.source_idx += target_slice_length


class JinjaAnalyzer:
    """Analyzes a Jinja template to prepare for tracing."""

    re_open_tag = regex.compile(r"^\s*({[{%])[\+\-]?\s*")
    re_close_tag = regex.compile(r"\s*[\+\-]?([}%]})\s*$")

    def __init__(self, raw_str: str, env: Environment):
        # Input
        self.raw_str: str = raw_str
        self.env = env

        # Output
        self.raw_sliced: List[RawFileSlice] = []
        self.raw_slice_info: Dict[RawFileSlice, RawSliceInfo] = {}
        self.sliced_file: List[TemplatedFileSlice] = []

        # Internal bookkeeping
        self.slice_id: int = 0
        # {% set %} or {% macro %} or {% call %}
        self.inside_set_macro_or_call: bool = False
        self.inside_block = False  # {% block %}
        self.stack: List[int] = []
        self.idx_raw: int = 0

    def next_slice_id(self) -> str:
        """Returns a new, unique slice ID."""
        result = "{0:#0{1}x}".format(self.slice_id, 34)[2:]
        self.slice_id += 1
        return result

    def slice_info_for_literal(self, length, prefix="") -> RawSliceInfo:
        """Returns a RawSliceInfo for a literal.

        In the alternate template, literals are replaced with a uniquely
        numbered, easy-to-parse literal. JinjaTracer uses this output as
        a "breadcrumb trail" to deduce the execution path through the template.

        This is important even if the original literal (i.e. in the raw SQL
        file) was empty, as is the case when Jinja whitespace control is used
        (e.g. "{%- endif -%}"), because fewer breadcrumbs means JinjaTracer has
        to *guess* the path, in which case it assumes simple, straight-line
        execution, which can easily be wrong with loops and conditionals.
        """
        unique_alternate_id = self.next_slice_id()
        alternate_code = f"\0{prefix}{unique_alternate_id}_{length}"
        return self.make_raw_slice_info(
            unique_alternate_id, alternate_code, inside_block=self.inside_block
        )

    def update_inside_set_call_macro_or_block(
        self,
        block_type: str,
        trimmed_parts: List[str],
        m_open: Optional[regex.Match],
        m_close: Optional[regex.Match],
        tag_contents: List[str],
    ) -> Tuple[Optional[RawSliceInfo], str]:
        """Based on block tag, update whether in a set/call/macro/block section."""
        if block_type == "block_start" and trimmed_parts[0] in (
            "block",
            "call",
            "macro",
            "set",
        ):
            # Jinja supports two forms of {% set %}:
            # - {% set variable = value %}
            # - {% set variable %}value{% endset %}
            # https://jinja.palletsprojects.com/en/2.10.x/templates/#block-assignments
            # When the second format is used, set one of the fields
            # 'inside_set_or_macro' or 'inside_block' to True. This info is
            # used elsewhere, as other code inside these regions require
            # special handling. (Generally speaking, JinjaAnalyzer ignores
            # the contents of these blocks, treating them like opaque templated
            # regions.)
            try:
                # Entering a set/macro block. Build a source string consisting
                # of just this one Jinja command and see if it parses. If so,
                # it's a standalone command. OTOH, if it fails with "Unexpected
                # end of template", it was the opening command for a block.
                self.env.from_string(
                    f"{self.env.block_start_string} {' '.join(trimmed_parts)} "
                    f"{self.env.block_end_string}"
                )
                # Here we should mutate the block type to just templated
                # so we don't treat it as a block.
                # https://github.com/sqlfluff/sqlfluff/issues/3750
                block_type = "templated"
            except TemplateSyntaxError as e:
                if (
                    isinstance(e.message, str)
                    and "Unexpected end of template" in e.message
                ):
                    # It was opening a block, thus we're inside a set, macro, or
                    # block.
                    if trimmed_parts[0] == "block":
                        self.inside_block = True
                    else:
                        result = None
                        if trimmed_parts[0] == "call":
                            assert m_open and m_close
                            result = self.track_call(m_open, m_close, tag_contents)
                        self.inside_set_macro_or_call = True
                        return result, block_type
                else:
                    raise  # pragma: no cover
        elif block_type == "block_end":
            if trimmed_parts[0] in ("endcall", "endmacro", "endset"):
                # Exiting a set or macro or block.
                self.inside_set_macro_or_call = False
            elif trimmed_parts[0] == "endblock":
                # Exiting a {% block %} block.
                self.inside_block = False
        return None, block_type

    def make_raw_slice_info(
        self,
        unique_alternate_id: Optional[str],
        alternate_code: Optional[str],
        inside_block: bool = False,
    ) -> RawSliceInfo:
        """Create RawSliceInfo as given, or "empty" if in set/macro block."""
        if not self.inside_set_macro_or_call:
            return RawSliceInfo(unique_alternate_id, alternate_code, [], inside_block)
        else:
            return RawSliceInfo(None, None, [], False)

    # We decide the "kind" of element we're dealing with using its _closing_
    # tag rather than its opening tag. The types here map back to similar types
    # of sections in the python slicer.
    block_types = {
        "variable_end": "templated",
        "block_end": "block",
        "comment_end": "comment",
        # Raw tags should behave like blocks. Note that
        # raw_end and raw_begin are whole tags rather
        # than blocks and comments where we get partial
        # tags.
        "raw_end": "block",
        "raw_begin": "block",
    }

    def analyze(self, make_template: Callable[[str], Template]) -> JinjaTracer:
        """Slice template in jinja."""
        # str_buff and str_parts are two ways we keep track of tokens received
        # from Jinja. str_buff concatenates them together, while str_parts
        # accumulates the individual strings. We generally prefer using
        # str_parts. That's because Jinja doesn't just split on whitespace, so
        # by keeping tokens as Jinja returns them, the code is more robust.
        # Consider the following:
        #   {% set col= "col1" %}
        # Note there's no space after col. Jinja splits this up for us. If we
        # simply concatenated the parts together and later split on whitespace,
        # we'd need some ugly, fragile logic to handle various whitespace
        # possibilities:
        #   {% set col= "col1" %}
        #   {% set col = "col1" %}
        #   {% set col ="col1" %}
        # By using str_parts and letting Jinja handle this, it just works.

        str_buff = ""
        str_parts = []

        # https://jinja.palletsprojects.com/en/2.11.x/api/#jinja2.Environment.lex
        block_idx = 0
        last_elem_type = None
        for _, elem_type, raw in self.env.lex(self.raw_str):
            if last_elem_type == "block_end" or elem_type == "block_start":
                block_idx += 1
            last_elem_type = elem_type

            if elem_type == "data":
                self.track_literal(raw, block_idx)
                continue
            str_buff += raw
            str_parts.append(raw)

            if elem_type.endswith("_begin"):
                self.handle_left_whitespace_stripping(raw, block_idx)

            raw_slice_info: RawSliceInfo = self.make_raw_slice_info(None, None)
            tag_contents = []
            # raw_end and raw_begin behave a little differently in
            # that the whole tag shows up in one go rather than getting
            # parts of the tag at a time.
            m_open = None
            m_close = None
            if elem_type.endswith("_end") or elem_type == "raw_begin":
                block_type = self.block_types[elem_type]
                block_subtype = None
                # Handle starts and ends of blocks
                if block_type in ("block", "templated"):
                    m_open = self.re_open_tag.search(str_parts[0])
                    m_close = self.re_close_tag.search(str_parts[-1])
                    if m_open and m_close:
                        tag_contents = self.extract_tag_contents(
                            str_parts, m_close, m_open, str_buff
                        )

                    if block_type == "block" and tag_contents:
                        block_type, block_subtype = self.extract_block_type(
                            tag_contents[0], block_subtype
                        )
                    if block_type == "templated" and tag_contents:
                        assert m_open and m_close
                        raw_slice_info = self.track_templated(
                            m_open, m_close, tag_contents
                        )
                (
                    raw_slice_info_temp,
                    block_type,
                ) = self.update_inside_set_call_macro_or_block(
                    block_type, tag_contents, m_open, m_close, tag_contents
                )
                if raw_slice_info_temp:
                    raw_slice_info = raw_slice_info_temp
                m_strip_right = regex.search(
                    r"\s+$", raw, regex.MULTILINE | regex.DOTALL
                )
                if elem_type.endswith("_end") and raw.startswith("-") and m_strip_right:
                    # Right whitespace was stripped after closing block. Split
                    # off the trailing whitespace into a separate slice. The
                    # desired behavior is to behave similarly as the left
                    # stripping case. Note that the stakes are a bit lower here,
                    # because lex() hasn't *omitted* any characters from the
                    # strings it returns, it has simply grouped them differently
                    # than we want.
                    trailing_chars = len(m_strip_right.group(0))
                    self.raw_sliced.append(
                        RawFileSlice(
                            str_buff[:-trailing_chars],
                            block_type,
                            self.idx_raw,
                            block_subtype,
                            block_idx,
                        )
                    )
                    self.raw_slice_info[self.raw_sliced[-1]] = raw_slice_info
                    slice_idx = len(self.raw_sliced) - 1
                    self.idx_raw += len(str_buff) - trailing_chars
                    self.raw_sliced.append(
                        RawFileSlice(
                            str_buff[-trailing_chars:],
                            "literal",
                            self.idx_raw,
                            None,
                            block_idx,
                        )
                    )
                    self.raw_slice_info[
                        self.raw_sliced[-1]
                    ] = self.slice_info_for_literal(0)
                    self.idx_raw += trailing_chars
                else:
                    self.raw_sliced.append(
                        RawFileSlice(
                            str_buff,
                            block_type,
                            self.idx_raw,
                            block_subtype,
                            block_idx,
                        )
                    )
                    self.raw_slice_info[self.raw_sliced[-1]] = raw_slice_info
                    slice_idx = len(self.raw_sliced) - 1
                    self.idx_raw += len(str_buff)
                if block_type.startswith("block"):
                    self.track_block_start(block_type, tag_contents[0])
                    self.track_block_end(block_type, tag_contents[0])
                    self.update_next_slice_indices(
                        slice_idx, block_type, tag_contents[0]
                    )
                str_buff = ""
                str_parts = []
        return JinjaTracer(
            self.raw_str,
            self.raw_sliced,
            self.raw_slice_info,
            self.sliced_file,
            make_template,
        )

    def track_templated(
        self, m_open: regex.Match, m_close: regex.Match, tag_contents: List[str]
    ) -> RawSliceInfo:
        """Compute tracking info for Jinja templated region, e.g. {{ foo }}."""
        unique_alternate_id = self.next_slice_id()
        open_ = m_open.group(1)
        close_ = m_close.group(1)
        # Here, we still need to evaluate the original tag contents, e.g. in
        # case it has intentional side effects, but also return a slice ID
        # for tracking.
        alternate_code = (
            f"\0{unique_alternate_id} {open_} " f"{''.join(tag_contents)} {close_}"
        )
        return self.make_raw_slice_info(unique_alternate_id, alternate_code)

    def track_call(
        self, m_open: regex.Match, m_close: regex.Match, tag_contents: List[str]
    ):
        """Set up tracking for "{% call ... %}"."""
        unique_alternate_id = self.next_slice_id()
        open_ = m_open.group(1)
        close_ = m_close.group(1)
        # Here, we still need to evaluate the original tag contents, e.g. in
        # case it has intentional side effects, but also return a slice ID
        # for tracking.
        alternate_code = (
            f"\0{unique_alternate_id} {open_} " f"{''.join(tag_contents)} {close_}"
        )
        return self.make_raw_slice_info(unique_alternate_id, alternate_code)

    def track_literal(self, raw: str, block_idx: int) -> None:
        """Set up tracking for a Jinja literal."""
        self.raw_sliced.append(
            RawFileSlice(
                raw,
                "literal",
                self.idx_raw,
                None,
                block_idx,
            )
        )
        # Replace literal text with a unique ID.
        self.raw_slice_info[self.raw_sliced[-1]] = self.slice_info_for_literal(
            len(raw), ""
        )
        self.idx_raw += len(raw)

    @staticmethod
    def extract_block_type(tag_name, block_subtype):
        """Determine block type."""
        # :TRICKY: Syntactically, the Jinja {% include %} directive looks like
        # a block, but its behavior is basically syntactic sugar for
        # {{ open("somefile).read() }}. Thus, treat it as templated code.
        if tag_name == "include":
            block_type = "templated"
        elif tag_name.startswith("end"):
            block_type = "block_end"
        elif tag_name.startswith("el"):
            # else, elif
            block_type = "block_mid"
        else:
            block_type = "block_start"
            if tag_name == "for":
                block_subtype = "loop"
        return block_type, block_subtype

    @staticmethod
    def extract_tag_contents(
        str_parts: List[str],
        m_close: regex.Match,
        m_open: regex.Match,
        str_buff: str,
    ) -> List[str]:
        """Given Jinja tag info, return the stuff inside the braces.

        I.e. Trim off the brackets and the whitespace.
        """
        if len(str_parts) >= 3:
            # Handle a tag received as individual parts.
            trimmed_parts = str_parts[1:-1]
            if trimmed_parts[0].isspace():
                del trimmed_parts[0]
            if trimmed_parts[-1].isspace():
                del trimmed_parts[-1]
        else:
            # Handle a tag received in one go.
            trimmed_content = str_buff[len(m_open.group(0)) : -len(m_close.group(0))]
            trimmed_parts = trimmed_content.split()
        return trimmed_parts

    def track_block_start(self, block_type: str, tag_name: str) -> None:
        """On starting a 'call' block, set slice_type to "templated"."""
        if block_type == "block_start" and tag_name == "call":
            # Replace RawSliceInfo for this slice with one that has block_type
            # "templated".
            old_raw_file_slice = self.raw_sliced[-1]
            self.raw_sliced[-1] = old_raw_file_slice._replace(slice_type="templated")

            # Move existing raw_slice_info entry since it's keyed by RawFileSlice.
            self.raw_slice_info[self.raw_sliced[-1]] = self.raw_slice_info[
                old_raw_file_slice
            ]
            del self.raw_slice_info[old_raw_file_slice]

    def track_block_end(self, block_type: str, tag_name: str) -> None:
        """On ending a 'for' or 'if' block, set up tracking."""
        if block_type == "block_end" and tag_name in (
            "endfor",
            "endif",
        ):
            # Replace RawSliceInfo for this slice with one that has alternate ID
            # and code for tracking. This ensures, for instance, that if a file
            # ends with "{% endif %} (with no newline following), that we still
            # generate a TemplateSliceInfo for it.
            unique_alternate_id = self.next_slice_id()
            alternate_code = f"{self.raw_sliced[-1].raw}\0{unique_alternate_id}_0"
            self.raw_slice_info[self.raw_sliced[-1]] = self.make_raw_slice_info(
                unique_alternate_id, alternate_code
            )

    def update_next_slice_indices(
        self, slice_idx: int, block_type: str, tag_name: str
    ) -> None:
        """Based on block, update conditional jump info."""
        if block_type == "block_start" and tag_name in (
            "for",
            "if",
        ):
            self.stack.append(slice_idx)
        elif block_type == "block_mid":
            # Record potential forward jump over this block.
            self.raw_slice_info[
                self.raw_sliced[self.stack[-1]]
            ].next_slice_indices.append(slice_idx)
            self.stack.pop()
            self.stack.append(slice_idx)
        elif block_type == "block_end" and tag_name in (
            "endfor",
            "endif",
        ):
            if not self.inside_set_macro_or_call:
                # Record potential forward jump over this block.
                self.raw_slice_info[
                    self.raw_sliced[self.stack[-1]]
                ].next_slice_indices.append(slice_idx)
                if self.raw_sliced[self.stack[-1]].slice_subtype == "loop":
                    # Record potential backward jump to the loop beginning.
                    self.raw_slice_info[
                        self.raw_sliced[slice_idx]
                    ].next_slice_indices.append(self.stack[-1] + 1)
                self.stack.pop()

    def handle_left_whitespace_stripping(self, token: str, block_idx: int) -> None:
        """If block open uses whitespace stripping, record it.

        When a "begin" tag (whether block, comment, or data) uses whitespace
        stripping
        (https://jinja.palletsprojects.com/en/3.0.x/templates/#whitespace-control)
        the Jinja lex() function handles this by discarding adjacent whitespace
        from 'raw_str'. For more insight, see the tokeniter() function in this file:
        https://github.com/pallets/jinja/blob/main/src/jinja2/lexer.py

        We want to detect and correct for this in order to:
        - Correctly update "idx" (if this is wrong, that's a potential
          DISASTER because lint fixes use this info to update the source file,
          and incorrect values often result in CORRUPTING the user's file so
          it's no longer valid SQL. :-O
        - Guarantee that the slices we return fully "cover" the contents of
          'in_str'.

        We detect skipped characters by looking ahead in in_str for the token
        just returned from lex(). The token text will either be at the current
        'idx_raw' position (if whitespace stripping did not occur) OR it'll be
        farther along in 'raw_str', but we're GUARANTEED that lex() only skips
        over WHITESPACE; nothing else.
        """
        # Find the token returned. Did lex() skip over any characters?
        num_chars_skipped = self.raw_str.index(token, self.idx_raw) - self.idx_raw
        if not num_chars_skipped:
            return

        # Yes. It skipped over some characters. Compute a string
        # containing the skipped characters.
        skipped_str = self.raw_str[self.idx_raw : self.idx_raw + num_chars_skipped]

        # Sanity check: Verify that Jinja only skips over
        # WHITESPACE, never anything else.
        if not skipped_str.isspace():  # pragma: no cover
            templater_logger.warning(
                "Jinja lex() skipped non-whitespace: %s", skipped_str
            )
        # Treat the skipped whitespace as a literal.
        self.raw_sliced.append(
            RawFileSlice(skipped_str, "literal", self.idx_raw, None, block_idx)
        )
        self.raw_slice_info[self.raw_sliced[-1]] = self.slice_info_for_literal(0)
        self.idx_raw += num_chars_skipped



"""String Helpers for the parser module."""

from typing import Iterator


def frame_msg(msg: str) -> str:
    """Frame a message with hashes so that it covers five lines."""
    return f"\n###\n#\n# {msg}\n#\n###"


def curtail_string(s: str, length=20) -> str:
    """Trim a string nicely to length."""
    if len(s) > length:
        return s[:length] + "..."
    else:
        return s


def findall(substr: str, in_str: str) -> Iterator[int]:
    """Yields all the positions sbstr within in_str.

    https://stackoverflow.com/questions/4664850/how-to-find-all-occurrences-of-a-substring
    """
    # Return nothing if one of the inputs is trivial
    if not substr or not in_str:
        return
    idx = in_str.find(substr)
    while idx != -1:
        yield idx
        idx = in_str.find(substr, idx + 1)



"""Implementation of Rule L019."""

from typing import List

from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.doc_decorators import (
    document_configuration,
    document_fix_compatible,
    document_groups,
)
from sqlfluff.utils.reflow import ReflowSequence


@document_groups
@document_fix_compatible
@document_configuration
class Rule_L019(BaseRule):
    """Leading/Trailing comma enforcement.

    **Anti-pattern**

    There is a mixture of leading and trailing commas.

    .. code-block:: sql

        SELECT
            a
            , b,
            c
        FROM foo

    **Best practice**

    By default, `SQLFluff` prefers trailing commas. However it
    is configurable for leading commas. The chosen style must be used
    consistently throughout your SQL.

    .. code-block:: sql

        SELECT
            a,
            b,
            c
        FROM foo

        -- Alternatively, set the configuration file to 'leading'
        -- and then the following would be acceptable:

        SELECT
            a
            , b
            , c
        FROM foo
    """

    groups = ("all",)
    crawl_behaviour = SegmentSeekerCrawler({"comma"})
    _adjust_anchors = True

    def _eval(self, context: RuleContext) -> List[LintResult]:
        """Enforce comma placement.

        For leading commas we're looking for trailing commas, so
        we look for newline segments. For trailing commas we're
        looking for leading commas, so we look for the comma itself.

        We also want to handle proper whitespace removal/addition. We remove
        any trailing whitespace after the leading comma, when converting a
        leading comma to a trailing comma. We add whitespace after the leading
        comma when converting a trailing comma to a leading comma.
        """
        return (
            ReflowSequence.from_around_target(
                context.segment,
                root_segment=context.parent_stack[0],
                config=context.config,
            )
            .rebreak()
            .get_results()
        )



"""Tests specific to the soql dialect."""
import pytest

from sqlfluff.core import FluffConfig, Linter
from sqlfluff.core.errors import SQLParseError


@pytest.mark.parametrize(
    "raw",
    [
        "ALTER TABLE foo DROP COLUMN bar\n",
        "CREATE USER my_user\n",
        "TRUNCATE TABLE foo\n",
        "EXPLAIN SELECT Id FROM Contact\n",
        "DROP TABLE foo\n",
        "DROP USER my_user\n",
    ],
)
def test_non_selects_unparseable(raw: str) -> None:
    """Test that non-SELECT commands are not parseable."""
    cfg = FluffConfig(configs={"core": {"dialect": "soql"}})
    lnt = Linter(config=cfg)
    result = lnt.lint_string(raw)
    assert len(result.violations) == 1
    assert isinstance(result.violations[0], SQLParseError)



"""Tests the python routines within L007."""

import sqlfluff
from sqlfluff.core.config import FluffConfig
from sqlfluff.core import Linter

EXPECTED_LEADING_MESSAGE = (
    "Found trailing binary operator. Expected only leading near line breaks."
)
EXPECTED_TRAILING_MESSAGE = (
    "Found leading binary operator. Expected only trailing near line breaks."
)


def test__rules__std_L007_default():
    """Verify that L007 returns the correct error message for default (trailing)."""
    sql = """
        SELECT
            a,
            b
        FROM foo
        WHERE
            a = 1 AND
            b = 2
    """
    result = sqlfluff.lint(sql)
    assert "L007" in [r["code"] for r in result]
    assert EXPECTED_LEADING_MESSAGE in [r["description"] for r in result]


def test__rules__std_L007_leading():
    """Verify correct error message when leading is used."""
    sql = """
        SELECT
            a,
            b
        FROM foo
        WHERE
            a = 1 AND
            b = 2
    """
    config = FluffConfig(
        configs={"layout": {"type": {"binary_operator": {"line_position": "leading"}}}},
        overrides={"dialect": "ansi"},
    )
    # The sqlfluff.lint API doesn't allow us to pass config so need to do what it does
    linter = Linter(config=config)
    result_records = linter.lint_string_wrapped(sql).as_records()
    result = result_records[0]["violations"]
    assert "L007" in [r["code"] for r in result]
    assert EXPECTED_LEADING_MESSAGE in [r["description"] for r in result]


def test__rules__std_L007_trailing():
    """Verify correct error message when trailing is used."""
    sql = """
        SELECT
            a,
            b
        FROM foo
        WHERE
            a = 1
            AND b = 2
    """
    config = FluffConfig(
        configs={
            "layout": {"type": {"binary_operator": {"line_position": "trailing"}}}
        },
        overrides={"dialect": "ansi"},
    )
    # The sqlfluff.lint API doesn't allow us to pass config so need to do what it does
    linter = Linter(config=config)
    result_records = linter.lint_string_wrapped(sql).as_records()
    result = result_records[0]["violations"]
    assert "L007" in [r["code"] for r in result]
    assert EXPECTED_TRAILING_MESSAGE in [r["description"] for r in result]



"""Testing utils for rule plugins."""
from sqlfluff.core import Linter
from sqlfluff.core.errors import SQLParseError, SQLTemplaterError
from sqlfluff.core.rules import get_ruleset
from sqlfluff.core.config import FluffConfig
from typing import Tuple, List, NamedTuple, Optional, Set
from glob import glob

import pytest
import yaml


class RuleTestCase(NamedTuple):
    """Used like a dataclass by rule tests."""

    rule: Optional[str] = None
    desc: Optional[str] = None
    pass_str: Optional[str] = None
    fail_str: Optional[str] = None
    violations: Optional[Set[dict]] = None
    fix_str: Optional[str] = None
    violations_after_fix: Optional[Set[dict]] = None
    configs: Optional[dict] = None
    skip: Optional[str] = None
    line_numbers: List[int] = []


def load_test_cases(
    test_cases_path: str,
) -> Tuple[List[str], List[RuleTestCase]]:
    """Load rule test cases from YAML files."""
    ids = []
    test_cases = []

    for path in sorted(glob(test_cases_path)):
        with open(path) as f:
            raw = f.read()

        y = yaml.safe_load(raw)

        rule = y.pop("rule")
        global_config = y.pop("configs", None)
        if global_config:
            for i in y:
                if not ("configs" in y[i].keys()):
                    y[i].update({"configs": global_config})
        ids.extend([rule + "_" + t for t in y])
        test_cases.extend([RuleTestCase(rule=rule, **v) for k, v in y.items()])

    return ids, test_cases


def get_rule_from_set(code, config):
    """Fetch a rule from the rule set."""
    for r in get_ruleset().get_rulelist(config=config):
        if r.code == code:  # pragma: no cover
            return r
    raise ValueError(f"{code!r} not in {get_ruleset()!r}")


def assert_rule_fail_in_sql(code, sql, configs=None, line_numbers=None):
    """Assert that a given rule does fail on the given sql."""
    print("# Asserting Rule Fail in SQL")
    # Set up the config to only use the rule we are testing.
    overrides = {"rules": code}
    if configs is None or "core" not in configs or "dialect" not in configs["core"]:
        overrides["dialect"] = "ansi"
    cfg = FluffConfig(configs=configs, overrides=overrides)
    # Lint it using the current config (while in fix mode)
    linted = Linter(config=cfg).lint_string(sql, fix=True)
    lerrs = linted.get_violations()
    print("Errors Found:")
    for e in lerrs:
        print("    " + repr(e))
        if e.desc().startswith("Unexpected exception"):
            pytest.fail(f"Linter failed with {e.desc()}")  # pragma: no cover
    parse_errors = list(
        filter(lambda v: isinstance(v, (SQLParseError, SQLTemplaterError)), lerrs)
    )
    if parse_errors:
        pytest.fail(f"Found the following parse errors in test case: {parse_errors}")
    if not any(v.rule.code == code for v in lerrs):
        print(f"Parsed File:\n{linted.tree.stringify()}")
        pytest.fail(
            f"No {code} failures found in query which should fail.",
            pytrace=False,
        )
    if line_numbers:
        actual_line_numbers = [e.line_no for e in lerrs]
        if line_numbers != actual_line_numbers:  # pragma: no cover
            pytest.fail(
                "Expected errors on lines {}, but got errors on lines {}".format(
                    line_numbers, actual_line_numbers
                )
            )
    fixed, _ = linted.fix_string()
    return fixed, linted.violations


def assert_rule_pass_in_sql(code, sql, configs=None, msg=None):
    """Assert that a given rule doesn't fail on the given sql."""
    # Configs allows overrides if we want to use them.
    print("# Asserting Rule Pass in SQL")
    if configs is None:
        configs = {}
    core = configs.setdefault("core", {})
    core["rules"] = code
    overrides = {}
    if "dialect" not in configs["core"]:
        overrides["dialect"] = "ansi"
    cfg = FluffConfig(configs=configs, overrides=overrides)
    linter = Linter(config=cfg)

    # This section is mainly for aid in debugging.
    rendered = linter.render_string(sql, fname="<STR>", config=cfg, encoding="utf-8")
    parsed = linter.parse_rendered(rendered, recurse=True)
    if parsed.violations:
        if msg:
            print(msg)  # pragma: no cover
        pytest.fail(parsed.violations[0].desc() + "\n" + parsed.tree.stringify())
    print(f"Parsed:\n {parsed.tree.stringify()}")

    # Note that lint_string() runs the templater and parser again, in order to
    # test the whole linting pipeline in the same way that users do. In other
    # words, the "rendered" and "parsed" variables above are irrelevant to this
    # line of code.
    lint_result = linter.lint_string(sql, config=cfg, fname="<STR>")
    lerrs = lint_result.violations
    if any(v.rule.code == code for v in lerrs):
        print("Errors Found:")
        for e in lerrs:
            print("    " + repr(e))

        if msg:
            print(msg)  # pragma: no cover
        pytest.fail(f"Found {code} failures in query which should pass.", pytrace=False)


def assert_rule_raises_violations_in_file(rule, fpath, violations, fluff_config):
    """Assert that a given rule raises given errors in specific positions of a file."""
    lntr = Linter(config=fluff_config)
    lnt = lntr.lint_path(fpath)
    # Reformat the test data to match the format we're expecting. We use
    # sets because we really don't care about order and if one is missing,
    # we don't care about the orders of the correct ones.
    assert set(lnt.check_tuples()) == {(rule, v[0], v[1]) for v in violations}


def prep_violations(rule, violations):
    """Default to test rule if code is omitted."""
    for v in violations:
        if "code" not in v:
            v["code"] = rule
    return violations


def assert_violations_before_fix(test_case, violations_before_fix):
    """Assert that the given violations are found in the given sql."""
    print("# Asserting Violations Before Fix")
    violation_info = [e.get_info_dict() for e in violations_before_fix]
    try:
        assert violation_info == prep_violations(test_case.rule, test_case.violations)
    except AssertionError:  # pragma: no cover
        print("Actual violations:\n" + yaml.dump(violation_info))
        raise


def assert_violations_after_fix(test_case):
    """Assert that the given violations are found in the fixed sql."""
    print("# Asserting Violations After Fix")
    _, violations_after_fix = assert_rule_fail_in_sql(
        test_case.rule,
        test_case.fix_str,
        configs=test_case.configs,
        line_numbers=test_case.line_numbers,
    )
    violation_info = [e.get_info_dict() for e in violations_after_fix]
    try:
        assert violation_info == prep_violations(
            test_case.rule, test_case.violations_after_fix
        )
    except AssertionError:  # pragma: no cover
        print("Actual violations_after_fix:\n" + yaml.dump(violation_info))
        raise


def rules__test_helper(test_case):
    """Test that a rule passes/fails on a set of test_cases.

    Optionally, also test the fixed string if provided in the test case.
    """
    if test_case.skip:
        pytest.skip(test_case.skip)

    if test_case.pass_str:
        assert_rule_pass_in_sql(
            test_case.rule,
            test_case.pass_str,
            configs=test_case.configs,
        )
    if test_case.fail_str:
        res, violations_before_fix = assert_rule_fail_in_sql(
            test_case.rule,
            test_case.fail_str,
            configs=test_case.configs,
            line_numbers=test_case.line_numbers,
        )
        if test_case.violations:
            assert_violations_before_fix(test_case, violations_before_fix)
        # If a `fixed` value is provided then check it matches
        if test_case.fix_str:
            assert res == test_case.fix_str
            if test_case.violations_after_fix:
                assert_violations_after_fix(test_case)
            else:
                assert_rule_pass_in_sql(
                    test_case.rule,
                    test_case.fix_str,
                    configs=test_case.configs,
                    msg="The SQL after fix is applied still contains rule violations. "
                    "To accept a partial fix, violations_after_fix must be set "
                    "listing the remaining, expected, violations.",
                )
        else:
            # Check that tests without a fix_str do not apply any fixes.
            assert res == test_case.fail_str, (
                "No fix_str was provided, but the rule modified the SQL. Where a fix "
                "can be applied by a rule, a fix_str must be supplied in the test."
            )



"""Helper utilities for reflow."""

from itertools import chain
from typing import Iterable, List

from sqlfluff.core.rules.base import LintFix, LintResult
from sqlfluff.core.parser import BaseSegment


def fixes_from_results(results: Iterable[LintResult]) -> List[LintFix]:
    """Return a list of fixes from an iterable of LintResult."""
    return list(chain.from_iterable(result.fixes for result in results))


def pretty_segment_name(segment: BaseSegment) -> str:
    """Get a nicely formatted name of the segment."""
    if segment.is_type("symbol"):
        # In a symbol reference, show the raw value and type.
        # (With underscores as spaces)
        return segment.get_type().replace("_", " ") + f" {segment.raw!r}"
    elif segment.is_type("keyword"):
        # Reference keywords as keywords.
        return f"{segment.raw!r} keyword"
    else:
        # Reference other segments just by their type.
        # (With underscores as spaces)
        return segment.get_type().replace("_", " ")



#!/usr/bin/env python

"""Utility strings for use during deployment.

NB: This is not part of the core sqlfluff code.
"""


# This contains various utility scripts

import shutil
import os
import click
import time
import subprocess
import sys
import yaml
import requests
import re
from ghapi.all import GhApi


@click.group()
def cli():
    """Launch the utility cli."""
    pass


@cli.command()
@click.option("--path", default=".test-reports")
def clean_tests(path):
    """Clear up the tests directory.

    NB: Using scripts allows platform independence
    Makes a new one afterward
    """
    try:
        shutil.rmtree(path)
        click.echo(f"Removed {path!r}...")
    # OSError is for python 27
    # in py36 its FileNotFoundError (but that inherits from IOError, which exists in
    # py27)
    except OSError:
        click.echo(f"Directory {path!r} does not exist. Skipping...")

    os.mkdir(path)
    click.echo(f"Created {path!r}")


@cli.command()
@click.argument("cmd", nargs=-1)
@click.option("--from-file", "-f", default=None)
@click.option("--runs", default=3, show_default=True)
def benchmark(cmd, runs, from_file):
    """Benchmark how long it takes to run a particular command."""
    if from_file:
        with open(from_file) as yaml_file:
            parsed = yaml.load(yaml_file.read(), Loader=yaml.FullLoader)
            benchmarks = parsed["benchmarks"]
            click.echo(repr(benchmarks))
    elif cmd:
        benchmarks = [{"name": str(hash(cmd)), "cmd": cmd}]
    else:
        click.echo("No command or file specified!")
        sys.exit(1)

    commit_hash = None
    post_results = False
    # Try and detect a CI environment
    if "CI" in os.environ:
        click.echo("CI detected!")
        commit_hash = os.environ.get("GITHUB_SHA", None)
        post_results = True
        click.echo(f"Commit hash is: {commit_hash!r}")

    all_results = {}
    for run_no in range(runs):
        click.echo(f"===== Run #{run_no + 1} =====")
        results = {}
        for benchmark in benchmarks:
            # Iterate through benchmarks
            click.echo("Starting benchmark: {!r}".format(benchmark["name"]))
            t0 = time.monotonic()
            click.echo("===START PROCESS OUTPUT===")
            process = subprocess.run(benchmark["cmd"])
            click.echo("===END PROCESS OUTPUT===")
            t1 = time.monotonic()
            if process.returncode != 0:
                if benchmark["cmd"][0] == "sqlfluff" and benchmark["cmd"][1] == "fix":
                    # Allow fix to fail as not all our benchmark errors are fixable
                    click.echo(
                        f"Fix command failed with return code: {process.returncode}"
                    )
                else:
                    click.echo(f"Command failed with return code: {process.returncode}")
                    sys.exit(process.returncode)
            else:
                duration = t1 - t0
                click.echo(f"Process completed in {duration:.4f}s")
                results[benchmark["name"]] = duration

        if post_results:
            click.echo(f"Posting results: {results}")
            api_key = os.environ["SQLFLUFF_BENCHMARK_API_KEY"]
            resp = requests.post(
                "https://f32cvv8yh3.execute-api.eu-west-1.amazonaws.com/result/gh"
                "/{repo}/{commit}".format(
                    # TODO: update the stats collector eventually to allow the new repo
                    # path
                    repo="alanmcruickshank/sqlfluff",
                    commit=commit_hash,
                ),
                params={"key": api_key},
                json=results,
            )
            click.echo(resp.text)
        all_results[run_no] = results
    click.echo("===== Done =====")
    for run_no in all_results:
        click.echo("Run {:>5}: {}".format(f"#{run_no}", all_results[run_no]))


@cli.command()
@click.option("--new_version_num")
def prepare_release(new_version_num):
    """Change version number in the cfg files."""
    api = GhApi(
        owner=os.environ["GITHUB_REPOSITORY_OWNER"],
        repo="sqlfluff",
        token=os.environ["GITHUB_TOKEN"],
    )
    releases = api.repos.list_releases()

    latest_draft_release = None
    for rel in releases:
        if rel["draft"]:
            latest_draft_release = rel
            break

    if not latest_draft_release:
        raise ValueError("No draft release found!")

    # Linkify the PRs and authors
    draft_body_parts = latest_draft_release["body"].split("\n")
    potential_new_contributors = []
    for i, p in enumerate(draft_body_parts):
        draft_body_parts[i] = re.sub(
            r"\(#([0-9]*)\) @([^ ]*)$",
            r"[#\1](https://github.com/sqlfluff/sqlfluff/pull/\1) [@\2](https://github.com/\2)",  # noqa E501
            p,
        )
        new_contrib_string = re.sub(
            r".*\(#([0-9]*)\) @([^ ]*)$",
            r"* [@\2](https://github.com/\2) made their first contribution in [#\1](https://github.com/sqlfluff/sqlfluff/pull/\1)",  # noqa E501
            p,
        )
        if new_contrib_string.startswith("* "):
            new_contrib_name = re.sub(r"\* \[(.*?)\].*", r"\1", new_contrib_string)
            potential_new_contributors.append(
                {"name": new_contrib_name, "line": new_contrib_string}
            )
    whats_changed_text = "\n".join(draft_body_parts)

    # Find the first commit for each contributor in this release
    potential_new_contributors.reverse()
    seen_contributors = set()
    deduped_potential_new_contributors = []
    for c in potential_new_contributors:
        if c["name"] not in seen_contributors:
            seen_contributors.add(c["name"])
            deduped_potential_new_contributors.append(c)

    input_changelog = open("CHANGELOG.md", encoding="utf8").readlines()
    write_changelog = open("CHANGELOG.md", "w", encoding="utf8")
    for i, line in enumerate(input_changelog):
        write_changelog.write(line)
        if "DO NOT DELETE THIS LINE" in line:
            existing_entry_start = i + 2
            # If the release is already in the changelog, update it
            if f"## [{new_version_num}]" in input_changelog[existing_entry_start]:
                input_changelog[
                    existing_entry_start
                ] = f"## [{new_version_num}] - {time.strftime('%Y-%m-%d')}\n"

                # Delete the existing What’s Changed and New Contributors sections
                remaining_changelog = input_changelog[existing_entry_start:]
                existing_whats_changed_start = (
                    next(
                        j
                        for j, line in enumerate(remaining_changelog)
                        if line.startswith("## What’s Changed")
                    )
                    + existing_entry_start
                )
                existing_new_contributors_start = (
                    next(
                        j
                        for j, line in enumerate(remaining_changelog)
                        if line.startswith("## New Contributors")
                    )
                    + existing_entry_start
                )
                existing_new_contributors_length = (
                    next(
                        j
                        for j, line in enumerate(
                            input_changelog[existing_new_contributors_start:]
                        )
                        if line.startswith("## [")
                    )
                    - 1
                )

                del input_changelog[
                    existing_whats_changed_start : existing_new_contributors_start
                    + existing_new_contributors_length
                ]

                # Now that we've cleared the previous sections, we will accurately
                # find if contributors have been previously mentioned in the changelog
                new_contributor_lines = []
                input_changelog_str = "".join(
                    input_changelog[existing_whats_changed_start:]
                )
                for c in deduped_potential_new_contributors:
                    if c["name"] not in input_changelog_str:
                        new_contributor_lines.append(c["line"])
                input_changelog[existing_whats_changed_start] = (
                    whats_changed_text
                    + "\n\n## New Contributors\n"
                    + "\n".join(new_contributor_lines)
                    + "\n\n"
                )

            else:
                write_changelog.write(
                    f"\n## [{new_version_num}] - {time.strftime('%Y-%m-%d')}\n\n## Highlights\n\n"  # noqa E501
                )
                write_changelog.write(whats_changed_text)
                write_changelog.write("\n## New Contributors\n\n")
                # Ensure contributor names don't appear in input_changelog list
                new_contributor_lines = []
                input_changelog_str = "".join(input_changelog)
                for c in deduped_potential_new_contributors:
                    if c["name"] not in input_changelog_str:
                        new_contributor_lines.append(c["line"])
                write_changelog.write("\n".join(new_contributor_lines))
                write_changelog.write("\n")

    write_changelog.close()

    for filename in ["setup.cfg", "plugins/sqlfluff-templater-dbt/setup.cfg"]:
        input_file = open(filename, "r").readlines()
        write_file = open(filename, "w")
        for line in input_file:
            for key in ["stable_version", "version"]:
                if line.startswith(key):
                    line = f"{key} = {new_version_num}\n"
                    break
            if line.startswith("    sqlfluff=="):
                line = f"    sqlfluff=={new_version_num}\n"
            write_file.write(line)
        write_file.close()

    for filename in ["docs/source/gettingstarted.rst"]:
        input_file = open(filename, "r").readlines()
        write_file = open(filename, "w")
        change_next_line = False
        for line in input_file:
            if change_next_line:
                line = f"    {new_version_num}\n"
                change_next_line = False
            elif line.startswith("    $ sqlfluff version"):
                change_next_line = True
            write_file.write(line)
        write_file.close()


if __name__ == "__main__":
    cli()



"""The core elements of sqlfluff."""

import tblib.pickling_support

# Config objects
from sqlfluff.core.config import FluffConfig

# Public classes
from sqlfluff.core.linter import Linter
from sqlfluff.core.parser import Lexer, Parser

# Dialect introspection
from sqlfluff.core.dialects import dialect_selector, dialect_readout

# All of the errors.
from sqlfluff.core.errors import (
    SQLBaseError,
    SQLTemplaterError,
    SQLLexError,
    SQLParseError,
    SQLLintError,
    SQLFluffUserError,
)

# Timing objects
from sqlfluff.core.timing import TimingSummary

__all__ = (
    "FluffConfig",
    "Linter",
    "Lexer",
    "Parser",
    "dialect_selector",
    "dialect_readout",
    "SQLBaseError",
    "SQLTemplaterError",
    "SQLLexError",
    "SQLParseError",
    "SQLLintError",
    "SQLFluffUserError",
    "TimingSummary",
)

# This is for "sqlfluff lint" and "sqlfluff fix" multiprocessing (--processes)
# support. If an exception (i.e. runtime error) occurs in a worker process, we
# want to return the tracebook to the main process and report it there, as part
# of the normal output. However, anything returned from a multiprocessing.Pool
# worker must be serializable using "pickle". By default, Python traceback
# objects cannot be pickled. The tblib package addresses this limitation; we
# simply need to install it before creating the worker pool. See these links for
# additional context:
# https://pypi.org/project/tblib/
# https://stackoverflow.com/questions/6126007/python-getting-a-traceback-from-a-multiprocessing-process
tblib.pickling_support.install()



"""The clickhouse dialect.

https://clickhouse.com/
"""

from sqlfluff.core.dialects import load_raw_dialect
from sqlfluff.core.parser import (
    AnyNumberOf,
    AnySetOf,
    BaseSegment,
    Bracketed,
    Delimited,
    Matchable,
    OneOf,
    OptionallyBracketed,
    Ref,
    Sequence,
)
from sqlfluff.dialects import dialect_ansi as ansi
from sqlfluff.dialects.dialect_clickhouse_keywords import (
    UNRESERVED_KEYWORDS,
)

ansi_dialect = load_raw_dialect("ansi")

clickhouse_dialect = ansi_dialect.copy_as("clickhouse")
clickhouse_dialect.sets("unreserved_keywords").update(UNRESERVED_KEYWORDS)
clickhouse_dialect.sets("reserved_keywords").clear()


class CTEDefinitionSegment(ansi.CTEDefinitionSegment):
    """A CTE Definition from a WITH statement.

    Overridden from ANSI to allow expression CTEs.
    https://clickhouse.com/docs/en/sql-reference/statements/select/with/
    """

    type = "common_table_expression"
    match_grammar: Matchable = OneOf(
        Sequence(
            Ref("SingleIdentifierGrammar"),
            Ref("CTEColumnList", optional=True),
            "AS",
            Bracketed(
                # Ephemeral here to subdivide the query.
                Ref("SelectableGrammar", ephemeral_name="SelectableGrammar")
            ),
        ),
        Sequence(
            Ref("ExpressionSegment"),
            "AS",
            Ref("SingleIdentifierGrammar"),
        ),
    )


class FromExpressionElementSegment(ansi.FromExpressionElementSegment):
    """A table expression.

    Overridden from ANSI to allow FINAL modifier.
    https://clickhouse.com/docs/en/sql-reference/statements/select/from#final-modifier
    """

    type = "from_expression_element"
    match_grammar: Matchable = Sequence(
        Ref("PreTableFunctionKeywordsGrammar", optional=True),
        OptionallyBracketed(Ref("TableExpressionSegment")),
        Ref(
            "AliasExpressionSegment",
            exclude=OneOf(
                Ref("SamplingExpressionSegment"),
                Ref("JoinLikeClauseGrammar"),
                Ref.keyword("Final"),
            ),
            optional=True,
        ),
        Ref.keyword("FINAL", optional=True),
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/arrays#flattening_arrays
        Sequence("WITH", "OFFSET", Ref("AliasExpressionSegment"), optional=True),
        Ref("SamplingExpressionSegment", optional=True),
        Ref("PostTableExpressionGrammar", optional=True),
    )


class EngineFunctionSegment(BaseSegment):
    """A ClickHouse `ENGINE` clause function.

    With this segment we attempt to match all possible engines.
    """

    type = "engine_function"
    match_grammar: Matchable = Sequence(
        Sequence(
            Ref(
                "FunctionNameSegment",
                exclude=OneOf(
                    Ref("DatePartFunctionNameSegment"),
                    Ref("ValuesClauseSegment"),
                ),
            ),
            Bracketed(
                Ref(
                    "FunctionContentsGrammar",
                    # The brackets might be empty for some functions...
                    optional=True,
                    ephemeral_name="FunctionContentsGrammar",
                ),
                # Engine functions may omit brackets.
                optional=True,
            ),
        ),
    )


class EngineSegment(BaseSegment):
    """An `ENGINE` used in `CREATE TABLE`."""

    type = "engine"

    match_grammar = Sequence(
        "ENGINE",
        Ref("EqualsSegment"),
        Sequence(
            Ref("EngineFunctionSegment"),
            AnySetOf(
                Sequence(
                    "ORDER",
                    "BY",
                    OneOf(
                        Ref("BracketedColumnReferenceListGrammar"),
                        Ref("ColumnReferenceSegment"),
                    ),
                    optional=True,
                ),
                Sequence(
                    "PARTITION",
                    "BY",
                    Ref("ExpressionSegment"),
                    optional=True,
                ),
                Sequence(
                    "PRIMARY",
                    "KEY",
                    Ref("ExpressionSegment"),
                    optional=True,
                ),
                Sequence(
                    "SAMPLE",
                    "BY",
                    Ref("ExpressionSegment"),
                    optional=True,
                ),
                Sequence(
                    "SETTINGS",
                    Delimited(
                        AnyNumberOf(
                            Sequence(
                                Ref("NakedIdentifierSegment"),
                                Ref("EqualsSegment"),
                                OneOf(
                                    Ref("NumericLiteralSegment"),
                                    Ref("QuotedLiteralSegment"),
                                ),
                                optional=True,
                            ),
                        )
                    ),
                    optional=True,
                ),
            ),
        ),
    )


class ColumnTTLSegment(BaseSegment):
    """A TTL clause for columns as used in CREATE TABLE.

    Specified in
    https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree/#mergetree-column-ttl
    """

    type = "column_ttl_segment"

    match_grammar = Sequence(
        "TTL",
        Ref("ExpressionSegment"),
    )


class TableTTLSegment(BaseSegment):
    """A TTL clause for tables as used in CREATE TABLE.

    Specified in
    https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree/#mergetree-table-ttl
    """

    type = "table_ttl_segment"

    match_grammar = Sequence(
        "TTL",
        Delimited(
            Sequence(
                Ref("ExpressionSegment"),
                OneOf(
                    "DELETE",
                    Sequence(
                        "TO",
                        "VOLUME",
                        Ref("QuotedLiteralSegment"),
                    ),
                    Sequence(
                        "TO",
                        "DISK",
                        Ref("QuotedLiteralSegment"),
                    ),
                    optional=True,
                ),
                Ref("WhereClauseSegment", optional=True),
                Ref("GroupByClauseSegment", optional=True),
            )
        ),
    )


class ColumnConstraintSegment(BaseSegment):
    """ClickHouse specific column constraints.

    As specified in
    https://clickhouse.com/docs/en/sql-reference/statements/create/table#constraints
    """

    type = "column_constraint_segment"

    match_grammar = AnySetOf(
        Sequence(
            Sequence(
                "CONSTRAINT",
                Ref("ObjectReferenceSegment"),
                optional=True,
            ),
            OneOf(
                Sequence(Ref.keyword("NOT", optional=True), "NULL"),
                Sequence("CHECK", Bracketed(Ref("ExpressionSegment"))),
                Sequence(
                    OneOf(
                        "DEFAULT",
                        "MATERIALIZED",
                        "ALIAS",
                    ),
                    OneOf(
                        Ref("LiteralGrammar"),
                        Ref("FunctionSegment"),
                        Ref("BareFunctionSegment"),
                    ),
                ),
                Sequence(
                    "EPHEMERAL",
                    OneOf(
                        Ref("LiteralGrammar"),
                        Ref("FunctionSegment"),
                        Ref("BareFunctionSegment"),
                        optional=True,
                    ),
                ),
                Ref("PrimaryKeyGrammar"),
                Sequence(
                    "CODEC",
                    Ref("FunctionContentsGrammar"),
                    optional=True,
                ),
                Ref("ColumnTTLSegment"),
            ),
        )
    )


class ClusterReferenceSegment(ansi.ObjectReferenceSegment):
    """A reference to a cluster."""

    type = "cluster_reference"


class CreateTableStatementSegment(ansi.CreateTableStatementSegment):
    """A `CREATE TABLE` statement.

    As specified in
    https://clickhouse.com/docs/en/sql-reference/statements/create/table/
    """

    type = "create_table_statement"

    match_grammar: Matchable = Sequence(
        "CREATE",
        OneOf(
            Ref("OrReplaceGrammar"),
            Ref.keyword("TEMPORARY"),
            optional=True,
        ),
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        Sequence(
            "ON",
            "CLUSTER",
            Ref("ClusterReferenceSegment"),
            optional=True,
        ),
        OneOf(
            # CREATE TABLE (...):
            Sequence(
                Bracketed(
                    Delimited(
                        OneOf(
                            Ref("TableConstraintSegment"),
                            Ref("ColumnDefinitionSegment"),
                            Ref("ColumnConstraintSegment"),
                        ),
                    ),
                    # Column definition may be missing if using AS SELECT
                    optional=True,
                ),
                Ref("EngineSegment"),
                # CREATE TABLE (...) AS SELECT:
                Sequence(
                    "AS",
                    Ref("SelectableGrammar"),
                    optional=True,
                ),
            ),
            # CREATE TABLE AS other_table:
            Sequence(
                "AS",
                Ref("TableReferenceSegment"),
                Ref("EngineSegment", optional=True),
            ),
            # CREATE TABLE AS table_function():
            Sequence(
                "AS",
                Ref("FunctionSegment"),
            ),
        ),
        AnySetOf(
            Sequence(
                "COMMENT",
                Ref("SingleQuotedIdentifierSegment"),
            ),
            Ref("TableTTLSegment"),
            optional=True,
        ),
        Ref("TableEndClauseSegment", optional=True),
    )


class StatementSegment(ansi.StatementSegment):
    """Overriding StatementSegment to allow for additional segment parsing."""

    match_grammar = ansi.StatementSegment.match_grammar
    parse_grammar = ansi.StatementSegment.parse_grammar.copy(
        insert=[
            Ref("EngineSegment"),
            Ref("ClusterReferenceSegment"),
            Ref("ColumnTTLSegment"),
            Ref("TableTTLSegment"),
        ]
    )



"""Define RuleContext class."""


import pathlib

from typing import (
    Optional,
    Tuple,
    Any,
)

from dataclasses import dataclass, field
from sqlfluff.core.config import FluffConfig

from sqlfluff.core.parser import BaseSegment, RawSegment
from sqlfluff.core.dialects import Dialect
from sqlfluff.core.templaters.base import TemplatedFile


@dataclass
class RuleContext:
    """Class for holding the context passed to rule eval functions."""

    # These don't change within a file.
    dialect: Dialect
    fix: bool
    templated_file: Optional[TemplatedFile]
    path: Optional[pathlib.Path]
    config: FluffConfig

    # These change within a file.
    # segment: The segment in question
    segment: BaseSegment
    # parent_stack: A tuple of the path from the root to this segment.
    parent_stack: Tuple[BaseSegment, ...] = field(default=tuple())
    # raw_stack: All of the raw segments so far in the file
    raw_stack: Tuple[RawSegment, ...] = field(default=tuple())
    # memory: Arbitrary storage for the rule
    memory: Any = field(default_factory=dict)
    # segment_idx: The index of this segment in the parent
    segment_idx: int = field(default=0)

    @property
    def siblings_pre(self) -> Tuple[BaseSegment, ...]:  # pragma: no cover
        """Return sibling segments prior to self.segment."""
        if self.parent_stack:
            return self.parent_stack[-1].segments[: self.segment_idx]
        else:
            return tuple()

    @property
    def siblings_post(self) -> Tuple[BaseSegment, ...]:
        """Return sibling segments after self.segment."""
        if self.parent_stack:
            return self.parent_stack[-1].segments[self.segment_idx + 1 :]
        else:
            return tuple()  # pragma: no cover



"""Implementation of Rule L064."""

from typing import Optional

import regex

from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.doc_decorators import (
    document_configuration,
    document_fix_compatible,
    document_groups,
)
from sqlfluff.utils.functional import rsp, FunctionalContext
from sqlfluff.core.parser.markers import PositionMarker
from sqlfluff.dialects.dialect_ansi import LiteralSegment


@document_groups
@document_fix_compatible
@document_configuration
class Rule_L064(BaseRule):
    r"""Consistent usage of preferred quotes for quoted literals.

    Some databases allow quoted literals to use either single or double quotes.
    Prefer one type of quotes as specified in rule setting, falling back to
    alternate quotes to reduce the need for escapes.

    Dollar-quoted raw strings are excluded from this rule, as they are mostly used for
    literal UDF Body definitions.

    .. note::
       This rule only checks quoted literals and not quoted identifiers as they often
       cannot interchange single and double quotes

       This rule is only enabled for dialects that allow single *and* double quotes for
       quoted literals (currently ``bigquery``, ``hive``, ``mysql``, ``sparksql``).
       It can be enabled for other dialects with the ``force_enable = True`` flag.

    **Anti-pattern**

    .. code-block:: sql
       :force:

        select
            "abc",
            'abc',
            "\"",
            "abc" = 'abc'
        from foo

    **Best practice**

    Ensure all quoted literals use preferred quotes, unless escaping can be reduced by
    using alternate quotes.

    .. code-block:: sql
       :force:

        select
            "abc",
            "abc",
            '"',
            "abc" = "abc"
        from foo

    """

    groups = ("all",)
    config_keywords = ["preferred_quoted_literal_style", "force_enable"]
    crawl_behaviour = SegmentSeekerCrawler({"literal"})
    targets_templated = True
    _dialects_with_double_quoted_strings = [
        "bigquery",
        "hive",
        "mysql",
        "sparksql",
    ]

    _quotes_mapping = {
        "single_quotes": {
            "common_name": "single quotes",
            "preferred_quote_char": "'",
            "alternate_quote_char": '"',
        },
        "double_quotes": {
            "common_name": "double quotes",
            "preferred_quote_char": '"',
            "alternate_quote_char": "'",
        },
    }
    # BigQuery string prefix characters.
    _string_prefix_chars = "rbRB"

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        # Config type hints
        self.preferred_quoted_literal_style: str
        self.force_enable: bool

        # Only care about quoted literal segments.
        if not context.segment.is_type("quoted_literal"):
            return None

        if not (
            self.force_enable
            or context.dialect.name in self._dialects_with_double_quoted_strings
        ):
            return LintResult(memory=context.memory)

        # This rule can also cover quoted literals that are partially templated.
        # I.e. when the quotes characters are _not_ part of the template we can
        # meaningfully apply this rule.
        templated_raw_slices = FunctionalContext(context).segment.raw_slices.select(
            rsp.is_slice_type("templated")
        )
        for raw_slice in templated_raw_slices:
            pos_marker = context.segment.pos_marker
            # This is to make mypy happy.
            assert isinstance(pos_marker, PositionMarker)

            # Check whether the quote characters are inside the template.
            # For the leading quote we need to account for string prefix characters.
            leading_quote_inside_template = pos_marker.source_str()[:2].lstrip(
                self._string_prefix_chars
            )[0] not in ['"', "'"]
            trailing_quote_inside_template = pos_marker.source_str()[-1] not in [
                '"',
                "'",
            ]

            # quotes are not entirely outside of a template, nothing we can do
            if leading_quote_inside_template or trailing_quote_inside_template:
                return LintResult(memory=context.memory)

        # If quoting style is set to consistent we use the quoting style of the first
        # quoted_literal that we encounter.
        if self.preferred_quoted_literal_style == "consistent":
            memory = context.memory
            preferred_quoted_literal_style = memory.get(
                "preferred_quoted_literal_style"
            )

            if not preferred_quoted_literal_style:
                # Getting the quote from LAST character to be able to handle STRING
                # prefixes
                preferred_quoted_literal_style = (
                    "double_quotes"
                    if context.segment.raw[-1] == '"'
                    else "single_quotes"
                )
                memory[
                    "preferred_quoted_literal_style"
                ] = preferred_quoted_literal_style
                self.logger.debug(
                    "Preferred string quotes is set to `consistent`. Derived quoting "
                    "style %s from first quoted literal.",
                    preferred_quoted_literal_style,
                )
        else:
            preferred_quoted_literal_style = self.preferred_quoted_literal_style

        fixed_string = self._normalize_preferred_quoted_literal_style(
            context.segment.raw,
            preferred_quote_char=self._quotes_mapping[preferred_quoted_literal_style][
                "preferred_quote_char"
            ],
            alternate_quote_char=self._quotes_mapping[preferred_quoted_literal_style][
                "alternate_quote_char"
            ],
        )

        if fixed_string != context.segment.raw:
            return LintResult(
                anchor=context.segment,
                memory=context.memory,
                fixes=[
                    LintFix.replace(
                        context.segment,
                        [
                            LiteralSegment(
                                raw=fixed_string,
                                type="quoted_literal",
                            )
                        ],
                    )
                ],
                description=(
                    "Inconsistent use of preferred quote style '"
                    f"{self._quotes_mapping[preferred_quoted_literal_style]['common_name']}"  # noqa: E501
                    f"'. Use {fixed_string} instead of {context.segment.raw}."
                ),
            )

        return None

    # Code for preferred quoted_literal style was copied from Black string normalization
    # and adapted to our use-case.
    def _regex_sub_with_overlap(
        self, regex: regex.Pattern, replacement: str, original: str
    ) -> str:
        """Replace `regex` with `replacement` twice on `original`.

        This is used by string normalization to perform replaces on overlapping matches.

        Source:
        https://github.com/psf/black/blob/7f7673d941a947a8d392c8c0866d3d588affc174/src/black/strings.py#L23-L29
        """
        return regex.sub(replacement, regex.sub(replacement, original))

    def _normalize_preferred_quoted_literal_style(
        self, s: str, preferred_quote_char: str, alternate_quote_char: str
    ) -> str:
        """Prefer `preferred_quote_char` but only if it doesn't cause more escaping.

        Adds or removes backslashes as appropriate.

        Source:
        https://github.com/psf/black/blob/7f7673d941a947a8d392c8c0866d3d588affc174/src/black/strings.py#L167
        """
        value = s.lstrip(self._string_prefix_chars)

        if value[:3] == preferred_quote_char * 3:
            # In triple-quoted strings we are not replacing escaped quotes.
            # So nothing left to do and we can exit.
            return s
        elif value[0] == preferred_quote_char:
            # Quotes are alright already. But maybe we can remove some unnecessary
            # escapes or reduce the number of escapes using alternate_quote_char ?
            orig_quote = preferred_quote_char
            new_quote = alternate_quote_char
        elif value[:3] == alternate_quote_char * 3:
            orig_quote = alternate_quote_char * 3
            new_quote = preferred_quote_char * 3
        elif value[0] == alternate_quote_char:
            orig_quote = alternate_quote_char
            new_quote = preferred_quote_char
        else:
            self.logger.debug(
                "Found quoted string %s using neither preferred quote char %s "
                "nor alternate_quote_char %s. Skipping...",
                s,
                preferred_quote_char,
                alternate_quote_char,
            )
            return s

        first_quote_pos = s.find(orig_quote)
        prefix = s[:first_quote_pos]
        unescaped_new_quote = regex.compile(rf"(([^\\]|^)(\\\\)*){new_quote}")
        escaped_new_quote = regex.compile(rf"([^\\]|^)\\((?:\\\\)*){new_quote}")
        escaped_orig_quote = regex.compile(rf"([^\\]|^)\\((?:\\\\)*){orig_quote}")
        body = s[first_quote_pos + len(orig_quote) : -len(orig_quote)]

        if "r" in prefix.lower():
            if unescaped_new_quote.search(body):
                self.logger.debug(
                    "There's at least one unescaped new_quote in this raw string "
                    "so converting is impossible."
                )
                return s

            # Do not modify the body of raw strings by introducing or removing
            # backslashes as this changes the value of the raw string.
            new_body = body
        else:
            # remove unnecessary escapes
            new_body = self._regex_sub_with_overlap(
                escaped_new_quote, rf"\1\2{new_quote}", body
            )
            if body != new_body:
                # Consider the string without unnecessary escapes as the original
                self.logger.debug("Removing unnecessary escapes in %s.", body)
                body = new_body
                s = f"{prefix}{orig_quote}{body}{orig_quote}"
            new_body = self._regex_sub_with_overlap(
                escaped_orig_quote, rf"\1\2{orig_quote}", new_body
            )
            new_body = self._regex_sub_with_overlap(
                unescaped_new_quote, rf"\1\\{new_quote}", new_body
            )

        if (
            new_quote == 3 * preferred_quote_char
            and new_body[-1:] == preferred_quote_char
        ):
            # edge case: for example when converting quotes from '''a"'''
            # to """a\"""" the last " of the string body needs to be escaped.
            new_body = new_body[:-1] + f"\\{preferred_quote_char}"

        orig_escape_count = body.count("\\")
        new_escape_count = new_body.count("\\")
        if new_escape_count > orig_escape_count:
            self.logger.debug(
                "Changing quote style would introduce more escapes in the body. "
                "Before: %s After: %s . Skipping.",
                body,
                new_body,
            )
            return s  # Do not introduce more escaping

        if new_escape_count == orig_escape_count and orig_quote == preferred_quote_char:
            # Already using preferred_quote_char, and no escape benefit to changing
            return s

        return f"{prefix}{new_quote}{new_body}{new_quote}"
